from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.files import (
    archive_files as api_client_archive_files,
    create_file,
    get_file,
    list_files,
    patch_file,
    unarchive_files,
)
from benchling_api_client.v2.stable.models.file import File
from benchling_api_client.v2.stable.models.file_create import FileCreate
from benchling_api_client.v2.stable.models.file_update import FileUpdate
from benchling_api_client.v2.stable.models.files_archival_change import FilesArchivalChange
from benchling_api_client.v2.stable.models.files_archive import FilesArchive
from benchling_api_client.v2.stable.models.files_archive_reason import FilesArchiveReason
from benchling_api_client.v2.stable.models.files_paginated_list import FilesPaginatedList
from benchling_api_client.v2.stable.models.files_unarchive import FilesUnarchive
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.models import ListFilesSort
from benchling_sdk.services.v2.base_service import BaseService


class FileService(BaseService):
    """
    Files.

    Files are Benchling objects that represent files and their metadata. Compared to Blobs, which are used by
    most Benchling products for attachments, Files are primarily used in the Analysis and Connect product.

    See https://benchling.com/api/v2/reference#/Files
    """

    @api_method
    def archive_files(
        self, file_ids: Iterable[str], reason: FilesArchiveReason
    ) -> FilesArchivalChange:
        """
        Archive Files.

        See https://benchling.com/api/reference#/Files/archiveFiles
        """
        archive_request = FilesArchive(reason=reason, file_ids=list(file_ids))
        response = api_client_archive_files.sync_detailed(
            client=self.client,
            json_body=archive_request,
        )
        return model_from_detailed(response)

    @api_method
    def create(self, file: FileCreate) -> File:
        """
        Create a file.

        See https://benchling.com/api/v2/reference#/Files/createFile
        """
        response = create_file.sync_detailed(client=self.client, json_body=file)
        return model_from_detailed(response)

    @api_method
    def get_by_id(self, file_id: str) -> File:
        """
        Get a file.

        See https://benchling.com/api/v2/reference#/Files/getFile
        """
        response = get_file.sync_detailed(client=self.client, file_id=file_id)
        return model_from_detailed(response)

    @api_method
    def _files_page(
        self,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        sort: Optional[ListFilesSort] = ListFilesSort.MODIFIEDAT,
        archive_reason: Optional[str] = None,
        created_at: Optional[str] = None,
        creator_ids: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[str] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        namesany_ofcase_sensitive: Optional[str] = None,
        namesany_of: Optional[str] = None,
        origin_ids: Optional[str] = None,
        ids: Optional[str] = None,
        display_ids: Optional[str] = None,
        returning: Optional[str] = None,
    ) -> Response[FilesPaginatedList]:
        response = list_files.sync_detailed(
            client=self.client,
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            sort=none_as_unset(sort),
            archive_reason=none_as_unset(archive_reason),
            created_at=none_as_unset(created_at),
            creator_ids=none_as_unset(creator_ids),
            folder_id=none_as_unset(folder_id),
            mentioned_in=none_as_unset(mentioned_in),
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            namesany_ofcase_sensitive=none_as_unset(namesany_ofcase_sensitive),
            namesany_of=none_as_unset(namesany_of),
            origin_ids=none_as_unset(origin_ids),
            ids=none_as_unset(ids),
            display_ids=none_as_unset(display_ids),
            returning=none_as_unset(returning),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        *,
        page_size: Optional[int] = 50,
        sort: Optional[ListFilesSort] = ListFilesSort.MODIFIEDAT,
        archive_reason: Optional[str] = None,
        created_at: Optional[str] = None,
        creator_ids: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[str] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        namesany_ofcase_sensitive: Optional[str] = None,
        namesany_of: Optional[str] = None,
        origin_ids: Optional[str] = None,
        ids: Optional[str] = None,
        display_ids: Optional[str] = None,
        returning: Optional[str] = None,
    ) -> PageIterator[File]:
        """
        List Files.

        See https://benchling.com/api/v2/reference#/Files/listFiles
        """

        def api_call(next_token: NextToken) -> Response[FilesPaginatedList]:
            return self._files_page(
                page_size=page_size,
                next_token=next_token,
                sort=sort,
                archive_reason=archive_reason,
                created_at=created_at,
                creator_ids=creator_ids,
                folder_id=folder_id,
                mentioned_in=mentioned_in,
                modified_at=modified_at,
                name=name,
                name_includes=name_includes,
                namesany_ofcase_sensitive=namesany_ofcase_sensitive,
                namesany_of=namesany_of,
                origin_ids=origin_ids,
                ids=ids,
                display_ids=display_ids,
                returning=returning,
            )

        def results_extractor(body: FilesPaginatedList) -> Optional[List[File]]:
            return body.files

        return PageIterator(api_call, results_extractor)

    @api_method
    def update(self, file_id: str, file: FileUpdate) -> File:
        """
        Update a File.

        See https://benchling.com/api/reference#/Files/updateFile
        """
        response = patch_file.sync_detailed(client=self.client, file_id=file_id, json_body=file)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, file_ids: Iterable[str]) -> FilesArchivalChange:
        """
        Unarchive one or more Files.

        See https://benchling.com/api/reference#/Files/unarchiveFiles
        """
        unarchive_request = FilesUnarchive(file_ids=list(file_ids))
        response = unarchive_files.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)



from __future__ import annotations

from benchling_sdk.apps.status.types import ReferencedSessionLinkType


def ref(reference: ReferencedSessionLinkType) -> str:
    """
    Ref.

    Helper method for easily serializing a referenced object into a string embeddable in AppSessionMessageCreate
    content.

    Example:
        dna_sequence = benchling.dna_sequences.get_by_id("seq_1234")
        AppSessionMessageCreate(f"This is my DNA sequence {ref(dna_sequence)} for analysis")

    """
    return ref_by_id(reference.id)


def ref_by_id(reference_id: str) -> str:
    """
    Ref by ID.

    Helper method for serializing a reference to an object into a string embeddable in
    AppSessionMessageCreate content via the reference object's id. This is helpful when clients
    don't already have access to the referenced object.

    Example:
        dna_sequence_id: str = "seq_asQya4lk"
        AppSessionMessageCreate(f"This is my DNA sequence {ref_by_id(dna_sequence_id)} for analysis")

    """
    assert reference_id, "reference_id cannot be empty or None"
    return _encode_id(reference_id)


def _encode_id(id: str) -> str:
    """Not sure {} are possible in Benchling IDs, but the spec says we're escaping."""
    escaped_id = id.replace("{", "\\{").replace("}", "\\}")
    return f"{{id:{escaped_id}}}"



from benchling_api_client.v2.stable.api.audit import audit_log
from benchling_api_client.v2.stable.models.audit_log_export import AuditLogExport

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.task_helpers import TaskHelper
from benchling_sdk.models import ExportAuditLogAsyncTaskResponse
from benchling_sdk.services.v2.base_service import BaseService


class AuditService(BaseService):
    """
    Audit Service.

    Export audit log data for Benchling objects.

    https://benchling.com/api/reference#/Audit
    """

    @api_method
    def get_audit_log(
        self, object_id: str, export: AuditLogExport
    ) -> TaskHelper[ExportAuditLogAsyncTaskResponse]:
        """
        Export an audit log file for a Benchling object.

        This endpoint launches a long-running task and returns a
        :py:class:`benchling_sdk.helpers.task_helpers.TaskHelper`. On success, the task's
        response will be an :py:class:`.ExportAuditLogAsyncTaskResponse` containing a link to
        download the exported audit log file from Amazon S3.

        This endpoint is subject to a rate limit of 500 requests per hour, in conjunction with
        the global request rate limit. Export throughput will additionally be rate limited around
        the scale of 70,000 total audit events exported in csv format or 30,000 total audit events
        exported in pdf format per hour.

        Example of submitting an export request and then getting the download URL from
        the completed task:

            task = benchling.v2.stable.audit.get_audit_log(object_id, export)
            task_result = task.wait_for_response()
            url = task_result.download_url

        See https://benchling.com/api/v2-beta/reference#/Audit/auditLog
        """
        response = audit_log.sync_detailed(client=self.client, object_id=object_id, json_body=export)
        return self._task_helper_from_response(response, ExportAuditLogAsyncTaskResponse)



from typing import List, Optional

from benchling_api_client.v2.stable.api.custom_notations import list_custom_notations
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.models import CustomNotation, CustomNotationsPaginatedList
from benchling_sdk.services.v2.base_service import BaseService


class CustomNotationService(BaseService):
    """
    Custom Notations.

    Benchling allows users to configure their own fully-custom string representation formats for import/export
    of nucleotide sequences (including chemical modifications).

    See https://benchling.com/api/reference#/Custom%20Notations
    """

    @api_method
    def _custom_notations_page(
        self,
        *,
        page_size: Optional[int] = None,
        next_token: Optional[str] = None,
    ) -> Response[CustomNotationsPaginatedList]:
        response = list_custom_notations.sync_detailed(
            client=self.client,
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(self, *, page_size: Optional[int] = None) -> PageIterator[CustomNotation]:
        """
        List custom notations.

        List all available custom notations for specifying modified nucleotide sequences.

        See https://benchling.com/api/reference#/Custom%20Notations/listCustomNotations
        """

        def api_call(next_token: NextToken) -> Response[CustomNotationsPaginatedList]:
            return self._custom_notations_page(
                page_size=page_size,
                next_token=next_token,
            )

        def results_extractor(
            body: CustomNotationsPaginatedList,
        ) -> Optional[List[CustomNotation]]:
            return body.custom_notations

        return PageIterator(api_call, results_extractor)



from abc import ABC
from typing import Optional, Type, TYPE_CHECKING

from benchling_api_client.v2.stable.client import Client
from benchling_api_client.v2.types import Response

from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.retry_helpers import RetryStrategy

if TYPE_CHECKING:
    from benchling_sdk.helpers.task_helpers import TaskHelper


class BaseService(ABC):
    """Abstract class for Benchling functional namespaces."""

    _client: Client
    _retry_strategy: RetryStrategy

    def __init__(self, client: Client, retry_strategy: Optional[RetryStrategy] = None):
        """
        Initialize a service.

        :param client: Underlying generated Client.
        :param retry_strategy: Retry strategy for failed HTTP calls
        """
        self._client = client
        self._retry_strategy = retry_strategy or RetryStrategy()

    @property
    def client(self) -> Client:
        """Provide access to the underlying generated Benchling Client."""
        return self._client

    @property
    def retry_strategy(self) -> RetryStrategy:
        """Provide access to the underlying user-specified RetryStrategy."""
        return self._retry_strategy

    def _create_service(self, cls):
        """
        Instantiate a service that derives from BaseService, with the same properties as this service.

        Override this in alpha and beta services that use a client other than self._client.
        """
        return cls(self._client, self._retry_strategy)

    def _task_helper_from_response(self, response: Response, result_type: Type) -> "TaskHelper":
        from benchling_sdk.helpers.task_helpers import TaskHelper

        return TaskHelper(model_from_detailed(response), self.client, result_type)



from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.lab_automation import (
    archive_automation_output_processors,
    create_automation_output_processor,
    generate_input_with_automation_input_generator,
    get_automation_input_generator,
    get_automation_output_processor,
    get_lab_automation_transform,
    list_automation_output_processors,
    process_output_with_automation_output_processor,
    unarchive_automation_output_processors,
    update_automation_input_generator,
    update_automation_output_processor,
    update_lab_automation_transform,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.helpers.task_helpers import TaskHelper
from benchling_sdk.models import (
    AutomationInputGenerator,
    AutomationInputGeneratorUpdate,
    AutomationOutputProcessor,
    AutomationOutputProcessorArchivalChange,
    AutomationOutputProcessorCreate,
    AutomationOutputProcessorsArchive,
    AutomationOutputProcessorsArchiveReason,
    AutomationOutputProcessorsPaginatedList,
    AutomationOutputProcessorsUnarchive,
    AutomationOutputProcessorUpdate,
    LabAutomationTransform,
    LabAutomationTransformUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService


class LabAutomationService(BaseService):
    """
    Lab Automation.

    Lab Automation endpoints support integration with lab instruments, and liquid handlers to create samples or
    results, and capture transfers between containers at scale.
    See https://benchling.com/api/reference#/Lab%20Automation
    """

    @api_method
    def input_generator_by_id(self, input_generator_id: str) -> AutomationInputGenerator:
        """
        Get an Automation Input Generator.

        See https://benchling.com/api/reference#/Lab%20Automation/getAutomationInputGenerator
        """
        response = get_automation_input_generator.sync_detailed(
            client=self.client, input_generator_id=input_generator_id
        )
        return model_from_detailed(response)

    @api_method
    def update_input_generator(
        self, input_generator_id: str, file_id: Optional[str]
    ) -> AutomationInputGenerator:
        """
        Update an Automation Input Generator.

        See https://benchling.com/api/reference#/Lab%20Automation/updateAutomationInputGenerator
        """
        update = AutomationInputGeneratorUpdate(file_id=file_id)
        response = update_automation_input_generator.sync_detailed(
            client=self.client, input_generator_id=input_generator_id, json_body=update
        )
        return model_from_detailed(response)

    @api_method
    def output_processor_by_id(self, output_processor_id: str) -> AutomationOutputProcessor:
        """
        Get an Automation Output Processor.

        See https://benchling.com/api/reference#/Lab%20Automation/getAutomationOutputProcessor
        """
        response = get_automation_output_processor.sync_detailed(
            client=self.client, output_processor_id=output_processor_id
        )
        return model_from_detailed(response)

    @api_method
    def update_output_processor(self, output_processor_id: str, file_id: str) -> AutomationOutputProcessor:
        """
        Update an Automation Output Processor.

        See https://benchling.com/api/reference#/Lab%20Automation/updateAutomationOutputProcessor
        """
        update = AutomationOutputProcessorUpdate(file_id=file_id)
        response = update_automation_output_processor.sync_detailed(
            client=self.client, output_processor_id=output_processor_id, json_body=update
        )
        return model_from_detailed(response)

    @api_method
    def generate_input(self, input_generator_id: str) -> TaskHelper[AutomationInputGenerator]:
        """
        Generate input with an Automation Input Generator.

        See https://benchling.com/api/reference#/Lab%20Automation/generateInputWithAutomationInputGenerator
        """
        response = generate_input_with_automation_input_generator.sync_detailed(
            client=self.client, input_generator_id=input_generator_id
        )
        return self._task_helper_from_response(response, AutomationInputGenerator)

    @api_method
    def process_output(self, output_processor_id: str) -> TaskHelper[AutomationOutputProcessor]:
        """
        Process output with an Automation Output Processor.

        See https://benchling.com/api/reference#/Lab%20Automation/processOutputWithAutomationOutputProcessor
        """
        response = process_output_with_automation_output_processor.sync_detailed(
            client=self.client, output_processor_id=output_processor_id
        )
        return self._task_helper_from_response(response, AutomationOutputProcessor)

    @api_method
    def _automation_output_processors_page(
        self,
        assay_run_id: str,
        automation_file_config_name: Optional[str],
        archive_reason: Optional[str],
        modified_at: Optional[str],
        next_token: Optional[NextToken] = None,
    ) -> Response[AutomationOutputProcessorsPaginatedList]:
        response = list_automation_output_processors.sync_detailed(
            client=self.client,
            assay_run_id=assay_run_id,
            next_token=none_as_unset(next_token),
            automation_file_config_name=none_as_unset(automation_file_config_name),
            archive_reason=none_as_unset(archive_reason),
            modified_at=none_as_unset(modified_at),
        )
        raise_for_status(response)
        return response  # type: ignore

    def automation_output_processors(
        self,
        assay_run_id: str,
        automation_file_config_name: Optional[str] = None,
        archive_reason: Optional[str] = None,
        modified_at: Optional[str] = None,
    ) -> PageIterator[AutomationOutputProcessor]:
        """
        List non-empty Automation Output Processors.

        Only Automation Output Processors which have an attached file will be included.
        See https://benchling.com/api/reference#/Lab%20Automation/listAutomationOutputProcessors
        """

        def api_call(next_token: NextToken) -> Response[AutomationOutputProcessorsPaginatedList]:
            return self._automation_output_processors_page(
                assay_run_id=assay_run_id,
                next_token=next_token,
                automation_file_config_name=automation_file_config_name,
                archive_reason=archive_reason,
                modified_at=modified_at,
            )

        def results_extractor(
            body: AutomationOutputProcessorsPaginatedList,
        ) -> List[AutomationOutputProcessor]:
            return body.automation_output_processors

        return PageIterator(api_call, results_extractor)

    @api_method
    def create_output_processor(
        self, automation_output_processor: AutomationOutputProcessorCreate
    ) -> AutomationOutputProcessor:
        """
        Create an Automation Output Processor.

        See https://benchling.com/api/reference#/Lab%20Automation/createAutomationOutputProcessor
        """
        response = create_automation_output_processor.sync_detailed(
            client=self.client, json_body=automation_output_processor
        )
        return model_from_detailed(response)

    @api_method
    def archive_automation_output_processors(
        self, automation_output_processor_ids: Iterable[str], reason: AutomationOutputProcessorsArchiveReason
    ) -> AutomationOutputProcessorArchivalChange:
        """
        Archive Automation Output Processors.

        See https://benchling.com/api/reference#/Lab%20Automation/archiveAutomationOutputProcessors
        """
        archive_request = AutomationOutputProcessorsArchive(
            reason=reason, automation_output_processor_ids=list(automation_output_processor_ids)
        )
        response = archive_automation_output_processors.sync_detailed(
            client=self.client, json_body=archive_request
        )
        return model_from_detailed(response)

    @api_method
    def unarchive_automation_output_processors(
        self, automation_output_processor_ids: Iterable[str]
    ) -> AutomationOutputProcessorArchivalChange:
        """
        Unarchive Automation Output Processors.

        See https://benchling.com/api/reference#/Lab%20Automation/unarchiveAutomationOutputProcessors
        """
        unarchive_request = AutomationOutputProcessorsUnarchive(
            automation_output_processor_ids=list(automation_output_processor_ids)
        )
        response = unarchive_automation_output_processors.sync_detailed(
            client=self.client, json_body=unarchive_request
        )
        return model_from_detailed(response)

    # TODO Should we rename this model LabAutomationTransform?
    @api_method
    def get_transform_by_id(self, transform_id: str) -> LabAutomationTransform:
        """
        Get a Lab Automation Transform step.

        See https://benchling.com/api/reference#/Lab%20Automation/getLabAutomationTransform
        """
        response = get_lab_automation_transform.sync_detailed(client=self.client, transform_id=transform_id)
        return model_from_detailed(response)

    @api_method
    def update_transform(
        self, transform_id: str, update: LabAutomationTransformUpdate
    ) -> LabAutomationTransform:
        """
        Update a Lab Automation Transform step.

        See https://benchling.com/api/reference#/Lab%20Automation/patchLabAutomationTransform
        """
        response = update_lab_automation_transform.sync_detailed(
            client=self.client, transform_id=transform_id, json_body=update
        )
        return model_from_detailed(response)



from benchling_api_client.v2.alpha.api.assemblies import (
    create_and_finalize_assembly,
    get_assembly,
    validate_assembly,
)
from benchling_api_client.v2.alpha.models.assembly import Assembly
from benchling_api_client.v2.alpha.models.assembly_spec_shared import AssemblySpecShared
from benchling_api_client.v2.alpha.models.create_and_finalize_assembly_json_body import (
    CreateAndFinalizeAssemblyJsonBody,
)

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.task_helpers import TaskHelper
from benchling_sdk.models import AsyncTaskLink
from benchling_sdk.services.v2.base_service import BaseService


class V2AlphaAssemblyService(BaseService):
    """
    V2-Alpha Assemblies.

    In Benchling, Assemblies are records of a process in which many fragment sequences are
    assembled in silico to create new construct sequences.

    https://benchling.com/api/v2-alpha/reference#/Assemblies
    """

    @api_method
    def get_by_id(self, bulk_assembly_id: str) -> Assembly:
        """
        Get a bulk assembly by its API identifier.

        See https://benchling.com/api/v2-alpha/reference#/Assemblies/GetAssembly
        """
        response = get_assembly.sync_detailed(client=self.client, bulk_assembly_id=bulk_assembly_id)
        return model_from_detailed(response)

    @api_method
    def create_and_finalize(self, create: CreateAndFinalizeAssemblyJsonBody) -> TaskHelper[Assembly]:
        """
        Create and finalize a new bulk assembly in a single step.

        This endpoint launches a long-running task and returns a
        :py:class:`benchling_sdk.helpers.task_helpers.TaskHelper` for waiting on the task.
        On success, the task's response will be an :py:class:`.Assembly`.

        See https://benchling.com/api/v2-alpha/reference#/Assemblies/CreateAndFinalizeAssembly
        """
        response = create_and_finalize_assembly.sync_detailed(client=self.client, json_body=create)
        return self._task_helper_from_response(response, Assembly)

    @api_method
    def validate(self, assembly: AssemblySpecShared) -> AsyncTaskLink:
        """
        Validate an assembly prior to finalization to see if it will succeed.

        This endpoint launches a long-running task and returns the Task ID of the launched task.
        The task response contains a set of validation results.

        See https://benchling.com/api/v2-alpha/reference#/Assemblies/ValidateAssembly
        """
        response = validate_assembly.sync_detailed(client=self.client, json_body=assembly)
        return model_from_detailed(response)



from typing import TypeVar, Union

from benchling_api_client.v2.extensions import UnknownType

from benchling_sdk.models import (
    ButtonUiBlock,
    ButtonUiBlockCreate,
    ButtonUiBlockUpdate,
    ChipUiBlock,
    ChipUiBlockCreate,
    ChipUiBlockUpdate,
    DropdownMultiValueUiBlock,
    DropdownMultiValueUiBlockCreate,
    DropdownMultiValueUiBlockUpdate,
    DropdownUiBlock,
    DropdownUiBlockCreate,
    DropdownUiBlockUpdate,
    FileUploadUiBlock,
    FileUploadUiBlockCreate,
    FileUploadUiBlockUpdate,
    MarkdownUiBlock,
    MarkdownUiBlockCreate,
    MarkdownUiBlockUpdate,
    SearchInputMultiValueUiBlock,
    SearchInputMultiValueUiBlockCreate,
    SearchInputMultiValueUiBlockUpdate,
    SearchInputUiBlock,
    SearchInputUiBlockCreate,
    SearchInputUiBlockUpdate,
    SectionUiBlock,
    SectionUiBlockCreate,
    SectionUiBlockUpdate,
    SelectorInputMultiValueUiBlock,
    SelectorInputMultiValueUiBlockCreate,
    SelectorInputMultiValueUiBlockUpdate,
    SelectorInputUiBlock,
    SelectorInputUiBlockCreate,
    SelectorInputUiBlockUpdate,
    TableUiBlock,
    TableUiBlockCreate,
    TableUiBlockUpdate,
    TextInputUiBlock,
    TextInputUiBlockCreate,
    TextInputUiBlockUpdate,
)

UiBlock = Union[
    ButtonUiBlock,
    ChipUiBlock,
    DropdownMultiValueUiBlock,
    DropdownUiBlock,
    FileUploadUiBlock,
    MarkdownUiBlock,
    SearchInputMultiValueUiBlock,
    SearchInputUiBlock,
    SectionUiBlock,
    SelectorInputMultiValueUiBlock,
    SelectorInputUiBlock,
    TableUiBlock,
    TextInputUiBlock,
    UnknownType,
]

UiBlockType = TypeVar(
    "UiBlockType",
    bound=UiBlock,
)

_UiBlockCreate = Union[
    ButtonUiBlockCreate,
    ChipUiBlockCreate,
    DropdownMultiValueUiBlockCreate,
    DropdownUiBlockCreate,
    FileUploadUiBlockCreate,
    MarkdownUiBlockCreate,
    SearchInputMultiValueUiBlockCreate,
    SearchInputUiBlockCreate,
    SectionUiBlockCreate,
    SelectorInputMultiValueUiBlockCreate,
    SelectorInputUiBlockCreate,
    TableUiBlockCreate,
    TextInputUiBlockCreate,
    UnknownType,
]

_UiBlockUpdate = Union[
    ButtonUiBlockUpdate,
    ChipUiBlockUpdate,
    DropdownMultiValueUiBlockUpdate,
    DropdownUiBlockUpdate,
    FileUploadUiBlockUpdate,
    MarkdownUiBlockUpdate,
    SearchInputMultiValueUiBlockUpdate,
    SearchInputUiBlockUpdate,
    SectionUiBlockUpdate,
    SelectorInputMultiValueUiBlockUpdate,
    SelectorInputUiBlockUpdate,
    TableUiBlockUpdate,
    TextInputUiBlockUpdate,
    UnknownType,
]

_UI_BLOCK_MAPPINGS_CREATE = {
    ButtonUiBlock: ButtonUiBlockCreate,
    ChipUiBlock: ChipUiBlockCreate,
    DropdownMultiValueUiBlock: DropdownMultiValueUiBlockCreate,
    DropdownUiBlock: DropdownUiBlockCreate,
    FileUploadUiBlock: FileUploadUiBlockCreate,
    MarkdownUiBlock: MarkdownUiBlockCreate,
    SearchInputMultiValueUiBlock: SearchInputMultiValueUiBlockCreate,
    SearchInputUiBlock: SearchInputUiBlockCreate,
    SectionUiBlock: SectionUiBlockCreate,
    SelectorInputMultiValueUiBlock: SelectorInputMultiValueUiBlockCreate,
    SelectorInputUiBlock: SelectorInputUiBlockCreate,
    TableUiBlock: TableUiBlockCreate,
    TextInputUiBlock: TextInputUiBlockCreate,
}

_UI_BLOCK_MAPPINGS_UPDATE = {
    ButtonUiBlock: ButtonUiBlockUpdate,
    ChipUiBlock: ChipUiBlockUpdate,
    DropdownMultiValueUiBlock: DropdownMultiValueUiBlockUpdate,
    DropdownUiBlock: DropdownUiBlockUpdate,
    FileUploadUiBlock: FileUploadUiBlockUpdate,
    MarkdownUiBlock: MarkdownUiBlockUpdate,
    SearchInputMultiValueUiBlock: SearchInputMultiValueUiBlockUpdate,
    SearchInputUiBlock: SearchInputUiBlockUpdate,
    SectionUiBlock: SectionUiBlockUpdate,
    SelectorInputMultiValueUiBlock: SelectorInputMultiValueUiBlockUpdate,
    SelectorInputUiBlock: SelectorInputUiBlockUpdate,
    TableUiBlock: TableUiBlockUpdate,
    TextInputUiBlock: TextInputUiBlockUpdate,
}



from __future__ import annotations

import json
from typing import cast, Dict, Generic, List, Optional, Protocol, Set, Type, TypeVar, Union

from benchling_api_client.v2.extensions import UnknownType
from benchling_api_client.v2.stable.types import UNSET, Unset

from benchling_sdk.apps.canvas.errors import (
    DuplicateBlockIdError,
    MissingResourceIdError,
    NoMatchingBlocksError,
)
from benchling_sdk.apps.canvas.types import (
    _UI_BLOCK_MAPPINGS_CREATE,
    _UI_BLOCK_MAPPINGS_UPDATE,
    _UiBlockCreate,
    _UiBlockUpdate,
    UiBlock,
    UiBlockType,
)
from benchling_sdk.apps.types import JsonType
from benchling_sdk.models import (
    AppCanvas,
    AppCanvasApp,
    AppCanvasCreate,
    AppCanvasUpdate,
    DropdownMultiValueUiBlock,
    DropdownUiBlock,
    SearchInputMultiValueUiBlock,
    SearchInputUiBlock,
    SectionUiBlock,
    SelectorInputMultiValueUiBlock,
    SelectorInputUiBlock,
    TableUiBlock,
    TextInputUiBlock,
)

S = TypeVar("S", bound="FilteredCanvasBuilderBlockStream")


def _ui_block_to_create(block: UiBlock) -> _UiBlockCreate:
    # Rely on the fact that the read/write shapes are compatible, for now
    if isinstance(block, UnknownType):
        return block
    elif type(block) in _UI_BLOCK_MAPPINGS_CREATE:
        create_block_class = _UI_BLOCK_MAPPINGS_CREATE[type(block)]
        create_block = create_block_class.from_dict(block.to_dict())  # type: ignore
        if isinstance(block, SectionUiBlock) and create_block.children:
            create_block.children = [_ui_block_to_create(child) for child in create_block.children]
        return create_block
    # Allow the block to be serialized as-is rather than runtime error for the user on a type check
    return block  # type: ignore


def _ui_block_to_update(block: UiBlock) -> _UiBlockUpdate:
    # Rely on the fact that the read/write shapes are compatible, for now
    # Update is functionally the same as create at the moment but different for type safety
    # and reserved in case the shapes do diverge later
    if isinstance(block, UnknownType):
        return block
    elif type(block) in _UI_BLOCK_MAPPINGS_UPDATE:
        update_block_class = _UI_BLOCK_MAPPINGS_UPDATE[type(block)]
        update_block = update_block_class.from_dict(block.to_dict())  # type: ignore
        if isinstance(block, SectionUiBlock) and update_block.children:
            update_block.children = [_ui_block_to_update(child) for child in update_block.children]
        return update_block
    # Allow the block to be serialized as-is rather than runtime error for the user on a type check
    return block  # type: ignore


class CanvasBuilderUiBlock(Generic[UiBlockType]):
    """Internal UI block wrapper for CanvasBuilder."""

    _block: UiBlockType
    _builder: CanvasBuilder

    def __init__(self, block: UiBlockType, builder: CanvasBuilder):
        """Init CanvasBuilderUiBlock."""
        self._block = block
        self._builder = builder

    @classmethod
    def from_api_model(cls, block: UiBlockType, builder: CanvasBuilder) -> CanvasBuilderUiBlock[UiBlockType]:
        """Create a _CanvasBuilderUiBlock from an underlying API model."""
        return cls(block, builder)

    def to_api_model(self) -> UiBlockType:
        """Convert to the underlying API model."""
        return self._block

    def children(self) -> CanvasBuilderBlockStream:
        """
        Return children for blocks when applicable, such as for section blocks.

        If not applicable, returns an empty canvas block stream with no blocks.
        """
        model = self.to_api_model()
        if isinstance(model, SectionUiBlock):
            # MyPy can't recognize the type narrowing when we check .children below
            section_block = cast(SectionUiBlock, model)
            child_blocks = [
                CanvasBuilderUiBlock.from_api_model(block, self._builder) for block in section_block.children
            ]
            # Pass reference to parent block (self)
            return CanvasBuilderBlockStream(self._builder, child_blocks, child_blocks, self)
        return CanvasBuilderBlockStream(self._builder, [], [])

    def replace(self, new_blocks: List[UiBlock]) -> None:
        """Replace block with provided new_blocks."""
        parent = self._parent_block()
        model = cast(UiBlock, self.to_api_model())
        if parent:
            self.insert_after(new_blocks)
            # Keeps MyPy happy; SectionUiBlock is not valid for .children of SectionUiBlock
            assert not (isinstance(model, SectionUiBlock) or isinstance(model, TableUiBlock))
            parent.children.remove(model)
        else:
            self.insert_after(new_blocks)
            # noinspection PyProtectedMember
            self._builder._source_canvas.blocks.remove(model)

    def remove(self) -> None:
        """Remove block."""
        parent = self._parent_block()
        model = cast(UiBlock, self.to_api_model())
        if parent:
            # Keeps MyPy happy; SectionUiBlock is not valid for .children of SectionUiBlock
            assert not (isinstance(model, SectionUiBlock) or isinstance(model, TableUiBlock))
            parent.children.remove(model)
        else:
            # noinspection PyProtectedMember
            self._builder._source_canvas.blocks.remove(model)

    def insert_after(self, new_blocks: List[UiBlock]) -> None:
        """Insert new_blocks after block."""
        self._nested_insert(new_blocks, 1)

    def insert_before(self, new_blocks: List[UiBlock]) -> None:
        """Insert new_blocks before block."""
        self._nested_insert(new_blocks, 0)

    def _nested_insert(self, new_blocks: List[UiBlock], offset: int) -> None:
        """
        Nested insert.

        This is used to handle the case that an insert is being performed and it's contextual within
        another block.

        For instance, the children of a SectionUiBlock.
        """
        parent_block = self._parent_block()
        if parent_block:
            child_blocks = parent_block.children
            # Using list() to solve "List" is invariant creates a copy which means this stops working
            self._insert(new_blocks, offset, child_blocks, self.to_api_model())  # type: ignore
            parent_block.children = child_blocks
            parent_builder_block = CanvasBuilderUiBlock.from_api_model(parent_block, self._builder)
            parent_builder_block.insert_after([parent_block])
            # noinspection PyProtectedMember
            self._builder._source_canvas.blocks.remove(parent_builder_block.to_api_model())
        else:
            # noinspection PyProtectedMember
            self._insert(new_blocks, offset, self._builder._source_canvas.blocks, self.to_api_model())

    @staticmethod
    def _insert(new_blocks: List[UiBlock], offset: int, blocks: List[UiBlock], target_block: UiBlock) -> None:
        """Insert new_blocks before block as a side effect."""
        index = blocks.index(target_block)
        for count, new_block in enumerate(new_blocks):
            blocks.insert(index + count + offset, new_block)

    def _parent_block(self) -> Optional[SectionUiBlock]:
        # noinspection PyProtectedMember
        for parent_block in self._builder._source_canvas.blocks:
            if isinstance(parent_block, SectionUiBlock):
                if self.to_api_model() in parent_block.children:
                    return parent_block
        return None


class CanvasBuilderFilter(Protocol):
    """Callable protocol for specifying a predicate for filtering UiBlocks."""

    def __call__(self, block: UiBlockType) -> bool:
        """Return True if the UiBlock matches specified conditions."""
        pass


class FilteredCanvasBuilderBlockStream:
    """Filtered UI block list wrapper for CanvasBuilder."""

    _builder: CanvasBuilder
    _blocks: List[CanvasBuilderUiBlock]
    _selected_blocks: List[CanvasBuilderUiBlock]
    _cursor: int

    def __init__(
        self,
        builder: CanvasBuilder,
        blocks: List[CanvasBuilderUiBlock],
        selected_blocks: List[CanvasBuilderUiBlock],
    ):
        """Init FilteredCanvasBuilderBlockStream."""
        self._builder = builder
        self._blocks = blocks
        self._selected_blocks = selected_blocks

    def __iter__(self):
        self._cursor = 0
        return self

    def __next__(self):
        if self._cursor >= len(self._blocks):
            raise StopIteration
        block = self._blocks[self._cursor]
        self._cursor += 1
        return block

    @classmethod
    def from_builder(cls: Type[S], builder: CanvasBuilder) -> S:
        """
        From Builder.

        Instantiate a new _FilteredCanvasBuilderBlockStream from a CanvasBuilder.
        """
        # noinspection PyProtectedMember
        blocks = [
            CanvasBuilderUiBlock.from_api_model(block, builder)
            for block in builder._source_canvas.blocks
            if not isinstance(block, UnknownType)
        ]
        return cls(builder, blocks, blocks)

    def filter(self, filter_function: CanvasBuilderFilter) -> FilteredCanvasBuilderBlockStream:
        """
        Filter.

        Accept a predicate that evaluates if a UiBlock should be included in the result or not.
        Returns a new stream of blocks filtered to the predicate, which is further operable.
        """
        return CanvasBuilderBlockStream(
            self._builder,
            self._blocks,
            [block for block in self._selected_blocks if filter_function(block.to_api_model())],
        )

    def get_by_id(self, block_id: str) -> CanvasBuilderUiBlock:
        """
        Get a block by its id.

        Raises NoMatchingBlocksError if the block is not found. To match an id as an optional, use filter().
        """
        matched_block = self._block_by_id(block_id, self._blocks)
        if not matched_block:
            raise NoMatchingBlocksError(f'Could not find a block with id "{block_id}"')
        return matched_block

    def _block_by_id(
        self, block_id: str, blocks: List[CanvasBuilderUiBlock]
    ) -> Optional[CanvasBuilderUiBlock]:
        for block in blocks:
            api_block = block.to_api_model()
            if api_block.id == block_id:
                return block
            children = block.children()
            if children.count() > 0:
                child_block = self._block_by_id(block_id, list(children))
                if child_block:
                    return child_block
        return None

    def count(self) -> int:
        """Return a count of the elements in the list of blocks."""
        return len(self._selected_blocks)

    def first(self) -> CanvasBuilderUiBlock:
        """Return the first block in the list."""
        if len(self._selected_blocks) < 1:
            raise NoMatchingBlocksError
        return self._selected_blocks[0]

    def last(self) -> CanvasBuilderUiBlock:
        """Return the last block in the list."""
        if len(self._selected_blocks) < 1:
            raise NoMatchingBlocksError
        return self._selected_blocks[-1]

    def remove(self) -> None:
        """Remove blocks."""
        updated_blocks = [
            block.to_api_model() for block in self._blocks if block not in self._selected_blocks
        ]
        # noinspection PyProtectedMember
        self._builder._source_canvas.blocks = updated_blocks


class CanvasBuilderBlockStream(FilteredCanvasBuilderBlockStream):
    """
    Internal UI block list wrapper for CanvasBuilder.

    Possesses some additional operations unavailable to filtered block streams.
    """

    _parent: Optional[CanvasBuilderUiBlock]

    def __init__(
        self,
        builder: CanvasBuilder,
        blocks: List[CanvasBuilderUiBlock],
        selected_blocks: List[CanvasBuilderUiBlock],
        parent: Optional[CanvasBuilderUiBlock] = None,
    ):
        """Init CanvasBuilderBlockStream."""
        super().__init__(builder, blocks, selected_blocks)
        self._parent = parent

    def append(self, new_blocks: List[UiBlock]) -> None:
        """
        Append new_blocks to the end of list of blocks.

        Only operates on unfiltered block streams. This can be a list of blocks on a Canvas, or children on a
        block such as SectionUiBlock.
        """
        if self._parent:
            if self.count() > 0:
                self.last().insert_after(new_blocks)
            else:
                self._parent.to_api_model().children.extend(new_blocks)
        else:
            # noinspection PyProtectedMember
            self._builder._source_canvas.blocks.extend(new_blocks)


class CanvasBuilder:
    """
    Canvas Builder.

    This class provides methods to help developers effectively work with Canvas UI Blocks.
    Working with the underlying API models directly can be clunky, as there is no native way to easily find
    blocks based on their attributes, and then operate on them. Some blocks, like SectionUiBlock, have nested
    children blocks, further complicating the unassisted experience.

    The goal of CanvasBuilder is to accept an existing Canvas, and easily change its contents (blocks), then
    send a resulting AppCanvasCreate or AppCanvasUpdate model to the API.

    Some block operations include:
        - Get by Id
        - Filtering by predicate
        - Remove
        - Replace
        - Insert After
        - Insert Before
        - Append

    Sample usage:
        ```
        canvas: Canvas = benchling.apps.get_canvas_by_id("canvas_id")
        builder = CanvasBuilder.from_canvas(canvas).blocks
            .filter(lambda block: isinstance(block, TextInputUiBlock))
            .remove()
        updated_canvas = benchling.apps.update_canvas("canvas_id", builder.to_update())
        ```
    """

    _source_canvas: AppCanvas
    _resource_id: str | Unset

    def __init__(
        self,
        app_id: str,
        feature_id: str,
        resource_id: Optional[str] = None,
        enabled: bool = True,
        session_id: Optional[str] = None,
        blocks: Optional[List[UiBlock]] = None,
        data: Optional[JsonType] = None,
    ):
        """
        Init AppCanvas.

        Create a CanvasBuilder from scratch. Useful when a source AppCanvas is not already created.
        """
        self._source_canvas = AppCanvas(
            app=AppCanvasApp(id=app_id),
            feature_id=feature_id,
            enabled=enabled,
            session_id=session_id,
            blocks=blocks if blocks else [],
            data=json.dumps(data) if data is not None else None,
        )
        self._resource_id = resource_id if resource_id is not None else UNSET

    @property
    def resource_id(self):
        """Returns the underlying resource_id of this CanvasBuilder."""
        return self._resource_id if self._resource_id is not UNSET else None

    @classmethod
    def from_canvas(cls, canvas: AppCanvas) -> CanvasBuilder:
        """
        From Canvas.

        Create a CanvasBuilder from an existing canvas. Preferred when a canvas already exists.
        """
        return cls(
            app_id=canvas.app.id,
            feature_id=canvas.feature_id,
            resource_id=canvas.resource_id,
            enabled=canvas.enabled,
            session_id=canvas.session_id,
            blocks=canvas.blocks,
            data=json.loads(canvas.data) if isinstance(canvas.data, str) else None,
        )

    def _with_enabled(self, value: bool) -> CanvasBuilder:
        # Encapsulated method for toggling Canvas enabled state
        return CanvasBuilder(
            app_id=self._source_canvas.app.id,
            feature_id=self._source_canvas.feature_id,
            resource_id=self.resource_id,
            enabled=value,
            session_id=self._source_canvas.session_id,
            blocks=self._source_canvas.blocks,
            data=self.data_to_json(),
        )

    def with_enabled(self, enabled: bool = True) -> CanvasBuilder:
        """
        Return a new CanvasBuilder with the underlying canvas enabled set to the specified value.

        Specify `False` to disable the canvas.
        This does not call the API, it only assigns state in the CanvasBuilder.
        """
        return self._with_enabled(enabled)

    def with_blocks(self, new_blocks: List[UiBlock]) -> CanvasBuilder:
        """
        Return a new CanvasBuilder with the underlying blocks replaced.

        This does not call the API, it only assigns state in the CanvasBuilder.
        """
        return CanvasBuilder(
            app_id=self._source_canvas.app.id,
            feature_id=self._source_canvas.feature_id,
            resource_id=self.resource_id,
            enabled=self._source_canvas.enabled,
            session_id=self._source_canvas.session_id,
            blocks=new_blocks,
            data=self.data_to_json(),
        )

    def with_data(self, new_data: Optional[JsonType]) -> CanvasBuilder:
        """
        Return a new CanvasBuilder with the underlying data replaced.

        This does not call the API, it only assigns state in the CanvasBuilder.
        """
        return CanvasBuilder(
            app_id=self._source_canvas.app.id,
            feature_id=self._source_canvas.feature_id,
            resource_id=self.resource_id,
            enabled=self._source_canvas.enabled,
            session_id=self._source_canvas.session_id,
            blocks=self._source_canvas.blocks,
            data=new_data,
        )

    def with_session_id(self, session_id: Optional[str]) -> CanvasBuilder:
        """
        Return a new CanvasBuilder with an optional session_id set.

        This does not call the API, it only assigns state in the CanvasBuilder.
        """
        return CanvasBuilder(
            app_id=self._source_canvas.app.id,
            feature_id=self._source_canvas.feature_id,
            resource_id=self.resource_id,
            enabled=self._source_canvas.enabled,
            session_id=session_id,
            blocks=self._source_canvas.blocks,
            data=self.data_to_json(),
        )

    def inputs_to_dict(self) -> Dict[str, Union[str, List[str]]]:
        """
        Read Inputs to dict.

        Return a dictionary of {block_id: block_value} for all blocks on the canvas with input values.
        Includes blocks with a value that can be a str or multivalued. Excludes TableUiBlock.
        Blocks that only have read attributes are omitted.

        List of included blocks:
            DropdownMultiValueUiBlock
            DropdownUiBlock
            SearchInputMultiValueUiBlock
            SearchInputUiBlock
            SelectorInputMultiValueUiBlock
            SelectorInputUiBlock
            TextInputUiBlock

        Raise DuplicateBlockIdError if multiple blocks with the same id are found.
        """
        return self._values_from_blocks(
            self._source_canvas.blocks,
            included_classes={
                DropdownMultiValueUiBlock,
                DropdownUiBlock,
                SearchInputMultiValueUiBlock,
                SearchInputUiBlock,
                SelectorInputMultiValueUiBlock,
                SelectorInputUiBlock,
                TextInputUiBlock,
            },
        )

    def inputs_to_dict_single_value(self) -> Dict[str, str]:
        """
        Read Inputs to dict, but only for single-valued blocks.

        Return a dictionary of {block_id: block_value} for all blocks on the canvas with single input values.
        Blocks that only have read attributes are omitted. Excludes TableUiBlock.

        List of included blocks:
            DropdownUiBlock
            SearchInputUiBlock
            SelectorInputUiBlock
            TextInputUiBlock

        Raise DuplicateBlockIdError if multiple blocks with the same id are found.
        """
        return cast(
            Dict[str, str],
            self._values_from_blocks(
                self._source_canvas.blocks,
                included_classes={
                    DropdownUiBlock,
                    SearchInputUiBlock,
                    SelectorInputUiBlock,
                    TextInputUiBlock,
                },
            ),
        )

    def inputs_to_dict_multi_value(self) -> Dict[str, List[str]]:
        """
        Read Inputs to dict, but only for multivalued blocks.

        Return a dictionary of {block_id: block_value} for all blocks on the canvas with multivalued input values.
        Blocks that only have read attributes are omitted. Excludes TableUiBlock.

        List of included blocks:
            DropdownMultiValueUiBlock
            SearchInputMultiValueUiBlock
            SelectorInputMultiValueUiBlock

        Raise DuplicateBlockIdError if multiple blocks with the same id are found.
        """
        return cast(
            Dict[str, List[str]],
            self._values_from_blocks(
                self._source_canvas.blocks,
                included_classes={
                    DropdownMultiValueUiBlock,
                    SearchInputMultiValueUiBlock,
                    SelectorInputMultiValueUiBlock,
                },
            ),
        )

    def _values_from_blocks(
        self,
        blocks: List[UiBlock],
        existing_keys: Optional[List[str]] = None,
        included_classes: Optional[Set[Type[UiBlock]]] = None,
    ) -> Dict[str, Union[str, List[str]]]:
        existing_keys = existing_keys if existing_keys else []
        values: Dict[str, Union[str, List[str]]] = dict()
        for block in blocks:
            # When included_classes is None, include all blocks
            if not isinstance(block, UnknownType) and (
                included_classes is None
                or isinstance(block, SectionUiBlock)
                or _is_included_class(included_classes, block)
            ):
                if hasattr(block, "value"):
                    if block.id in existing_keys:
                        raise DuplicateBlockIdError(
                            f'More than one block with the id "{block.id}" already exists'
                        )
                    existing_keys.append(block.id)
                    # Ignore type error since we checked hasattr above
                    values[block.id] = block.value  # type: ignore
                elif isinstance(block, SectionUiBlock):
                    # Cast to appease MyPy
                    parent_block = cast(SectionUiBlock, block)
                    # list() prevents MyPy from complaining: "List" is invariant
                    child_values = self._values_from_blocks(
                        list(parent_block.children), existing_keys, included_classes=included_classes
                    )
                    values.update(child_values)
        return values

    def to_update(self) -> AppCanvasUpdate:
        """Return an AppCanvasUpdate API model from the current state of the canvas managed by CanvasBuilder."""
        return AppCanvasUpdate(
            feature_id=self._source_canvas.feature_id,
            resource_id=self._resource_id,
            enabled=self._source_canvas.enabled,
            session_id=self._source_canvas.session_id,
            blocks=[_ui_block_to_update(block) for block in self._source_canvas.blocks],
            data=self._source_canvas.data,
        )

    def to_create(self) -> AppCanvasCreate:
        """
        Return an AppCanvasCreate API model from the current state of the canvas managed by CanvasBuilder.

        Raise MissingResourceIdError if the current state of the canvas managed by CanvasBuilder does not include
        a resource ID. Without a resource ID indicating the Benchling object on which to place the new canvas,
        canvas creation is underspecified and cannot be completed successfully.
        """
        if self.resource_id is None:
            raise MissingResourceIdError

        return AppCanvasCreate(
            app_id=self._source_canvas.app.id,
            feature_id=self._source_canvas.feature_id,
            resource_id=self.resource_id,
            enabled=self._source_canvas.enabled,
            session_id=self._source_canvas.session_id,
            blocks=[_ui_block_to_create(block) for block in self._source_canvas.blocks],
            data=self._source_canvas.data,
        )

    @property
    def blocks(self) -> CanvasBuilderBlockStream:
        """
        Blocks.

        Return a stream of blocks which can be iterated and operated on to mutate the canvas
        stored by the builder.
        """
        return CanvasBuilderBlockStream.from_builder(self)

    def data_to_json(self) -> Optional[JsonType]:
        """
        Convert Canvas data to JSON.

        Return a JSON object parsed from the string of the canvas's `data`, if present. Otherwise, return None.
        """
        return json.loads(self._source_canvas.data) if self._source_canvas.data is not None else None

    def __eq__(self, other) -> bool:
        return isinstance(other, CanvasBuilder) and self._source_canvas == other._source_canvas


def _is_included_class(included_classes: Set[Type[UiBlock]], target_class: UiBlock) -> bool:
    return isinstance(target_class, tuple(c for c in included_classes))



from datetime import datetime
from enum import Enum
from functools import cache
from typing import cast, List, Optional, Union

from benchling_api_client.v2.beta.extensions import Enums
from benchling_api_client.v2.beta.models.base_manifest_config import BaseManifestConfig
from benchling_api_client.v2.beta.models.dropdown_dependency import DropdownDependency
from benchling_api_client.v2.beta.models.dropdown_dependency_types import DropdownDependencyTypes
from benchling_api_client.v2.beta.models.entity_schema_dependency import EntitySchemaDependency
from benchling_api_client.v2.beta.models.entity_schema_dependency_type import EntitySchemaDependencyType
from benchling_api_client.v2.beta.models.field_definitions_manifest import FieldDefinitionsManifest
from benchling_api_client.v2.beta.models.manifest_array_config import ManifestArrayConfig
from benchling_api_client.v2.beta.models.manifest_float_scalar_config import ManifestFloatScalarConfig
from benchling_api_client.v2.beta.models.manifest_integer_scalar_config import ManifestIntegerScalarConfig
from benchling_api_client.v2.beta.models.manifest_scalar_config import ManifestScalarConfig
from benchling_api_client.v2.beta.models.manifest_text_scalar_config import ManifestTextScalarConfig
from benchling_api_client.v2.beta.models.resource_dependency import ResourceDependency
from benchling_api_client.v2.beta.models.resource_dependency_types import ResourceDependencyTypes
from benchling_api_client.v2.beta.models.schema_dependency import SchemaDependency
from benchling_api_client.v2.beta.models.schema_dependency_subtypes import SchemaDependencySubtypes
from benchling_api_client.v2.beta.models.schema_dependency_types import SchemaDependencyTypes
from benchling_api_client.v2.beta.models.workflow_task_schema_dependency import WorkflowTaskSchemaDependency
from benchling_api_client.v2.beta.models.workflow_task_schema_dependency_output import (
    WorkflowTaskSchemaDependencyOutput,
)
from benchling_api_client.v2.beta.models.workflow_task_schema_dependency_type import (
    WorkflowTaskSchemaDependencyType,
)
from benchling_api_client.v2.extensions import UnknownType
from benchling_api_client.v2.stable.extensions import NotPresentError

_ArrayElementDependency = Union[
    SchemaDependency,
    EntitySchemaDependency,
    WorkflowTaskSchemaDependency,
    DropdownDependency,
    ResourceDependency,
    ManifestScalarConfig,
]


class _UnsupportedSubTypeError(Exception):
    """Error when an unsupported subtype is encountered."""

    pass


class _ScalarConfigTypes(Enums.KnownString):
    """
    Enum type copied from an earlier version of benchling-api-client, for internal use only.

    See BNCH-108704.
    """

    TEXT = "text"
    FLOAT = "float"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    DATE = "date"
    DATETIME = "datetime"
    SECURE_TEXT = "secure_text"
    JSON = "json"

    def __str__(self) -> str:
        return str(self.value)

    @staticmethod
    @cache
    def of_unknown(val: str) -> "_ScalarConfigTypes":
        if not isinstance(val, str):
            raise ValueError(f"Value of _ScalarConfigTypes must be a string (encountered: {val})")
        newcls = Enum("_ScalarConfigTypes", {"_UNKNOWN": val}, type=Enums.UnknownString)  # type: ignore
        return cast(_ScalarConfigTypes, newcls._UNKNOWN)


def _field_definitions_from_dependency(
    dependency: Union[
        EntitySchemaDependency,
        SchemaDependency,
        WorkflowTaskSchemaDependency,
        WorkflowTaskSchemaDependencyOutput,
    ]
) -> List[FieldDefinitionsManifest]:
    """Safely return a list of field definitions from a schema dependency or empty list."""
    try:
        if hasattr(dependency, "field_definitions"):
            return dependency.field_definitions
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return []


def _element_definition_from_dependency(dependency: ManifestArrayConfig) -> List[_ArrayElementDependency]:
    """Safely return an element definition as a list of dependencies from an array dependency or empty list."""
    try:
        if hasattr(dependency, "element_definition"):
            return [
                _fix_element_definition_deserialization(element) for element in dependency.element_definition
            ]
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return []


def _enum_from_dependency(
    dependency: Union[
        ManifestFloatScalarConfig,
        ManifestIntegerScalarConfig,
        ManifestTextScalarConfig,
    ]
) -> List:
    """Safely return an enum from a scalar config."""
    try:
        if hasattr(dependency, "enum"):
            return dependency.enum
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return []


# TODO BNCH-57036 All element definitions currently deserialize to UnknownType. Hack around this temporarily
def _fix_element_definition_deserialization(  # noqa:PLR0911
    element: Union[UnknownType, _ArrayElementDependency]
) -> _ArrayElementDependency:
    if isinstance(element, UnknownType):
        if "type" in element.value:
            element_type = element.value["type"]
            if element_type == WorkflowTaskSchemaDependencyType.WORKFLOW_TASK_SCHEMA:
                return WorkflowTaskSchemaDependency.from_dict(element.value)
            elif element_type == EntitySchemaDependencyType.ENTITY_SCHEMA:
                return EntitySchemaDependency.from_dict(element.value)
            elif element_type in [member.value for member in SchemaDependencyTypes]:
                return SchemaDependency.from_dict(element.value)
            elif element_type == DropdownDependencyTypes.DROPDOWN:
                return DropdownDependency.from_dict(element.value)
            elif element_type in [member.value for member in ResourceDependencyTypes]:
                return ResourceDependency.from_dict(element.value)
            elif element_type in [member.value for member in _ScalarConfigTypes]:
                return type(element_type).from_dict(element.value)
        raise NotImplementedError(f"No array deserialization fix for {element}")
    return element


def _workflow_task_schema_output_from_dependency(
    dependency: WorkflowTaskSchemaDependency,
) -> Optional[WorkflowTaskSchemaDependencyOutput]:
    """Safely return a workflow task schema output from a workflow task schema or None."""
    try:
        if hasattr(dependency, "output"):
            return dependency.output
    # We can't seem to handle this programmatically by checking isinstance() or output truthiness
    except NotPresentError:
        pass
    return None


def _options_from_dependency(dependency: DropdownDependency) -> List[BaseManifestConfig]:
    """Safely return a list of options from a dropdown dependency or empty list."""
    try:
        if hasattr(dependency, "options"):
            return dependency.options
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return []


def _subtype_from_entity_schema_dependency(
    dependency: EntitySchemaDependency,
) -> Optional[SchemaDependencySubtypes]:
    """Safely return an entity schema dependency's subtype, if present."""
    try:
        if hasattr(dependency, "subtype") and dependency.subtype:
            return dependency.subtype
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return None


def datetime_config_value_to_str(value: datetime) -> str:
    """Convert a datetime value to a valid string accepted by a datetime app config item."""
    return value.strftime("%Y-%m-%d %I:%M:%S %p")



import threading
from typing import Any, Iterable, Iterator, List, Optional, Protocol, TypeVar

from benchling_api_client.v2.types import Response

from benchling_sdk.helpers.logging_helpers import sdk_logger
from benchling_sdk.helpers.response_helpers import model_from_detailed

# TODO BNCH-15439 Make bounded if we can inherit from a common model
ResultsBody = TypeVar("ResultsBody")
Model = TypeVar("Model")
NextToken = str

RESULTS_COUNT_HEADER = "Result-Count"


class PagedApiCall(Protocol):
    """Make an HTTP call which implemented pagination."""

    def __call__(self, next_token: NextToken) -> Response[Any]:
        """Perform an HTTP call, passing the next token from the API."""
        pass


class ResultsExtractor(Protocol):
    """Extract results from a paginated response."""

    def __call__(self, body: Any) -> Optional[List[Any]]:
        """
        Extract a list of modeled results from an HTTP response.

        Typically this entails knowing the attribute in the response which contains the list.
        """
        pass


class TokenExtractor(Protocol):
    """Get the next token."""

    def __call__(self, body: Any) -> NextToken:
        """Extract an API pagination token from an HTTP response."""
        pass


class LengthExtractor(Protocol):
    """Extract estimated size information if available."""

    def __call__(self, response: Response[Model]) -> Optional[int]:
        """Calculate the estimated total size of paginated records."""
        pass


def _default_token_extractor(body: ResultsBody) -> NextToken:
    """
    Extract the `next_token` from a result.

    A default method that returns the `next_token` from a model returned by
    the API. Since the generated API models do not inherit from a common base class,
    the `ResultsBody` type is unbounded.

    :param body: A model returned from the API which should have a `next_token` attribute
    :return: A str of the `next_token` from the results body
    """
    return body.next_token  # type: ignore


def _default_length_extractor(response: Response[Model]) -> Optional[int]:
    """
    Extract the estimated size (length) of total results from a response.

    A default method that returns the estimated length of total available results from the API.
    It will return None if the API did not indicate how many results were available.

    :param response: A response returned from the API
    :return: An optional int containing the estimated length of the results
    """
    if (
        response.headers
        and RESULTS_COUNT_HEADER in response.headers
        and response.headers[RESULTS_COUNT_HEADER].isdigit()
    ):
        return int(response.headers[RESULTS_COUNT_HEADER])
    return None


class PageIterator(Iterable[List[Model]]):
    """
    Paginate arbitrary Benchling API endpoints which support the concept of a `next_token`.

    `next_token` points to subsequent pages of results. Supporting API
    endpoints typically return a "results body" object which contains a key for `next_token`,
    as well as a key for a collection of results representing the current page.

    Any API errors encountered during retrieval of a page will be marshaled to
    :class:`benchling_sdk.errors.BenchlingError`.
    """

    _api_call: PagedApiCall
    _results_extractor: ResultsExtractor
    _token_extractor: TokenExtractor
    _length_extractor: LengthExtractor
    _next_token: Optional[NextToken]
    _approximate_length: Optional[int]
    _first_page: Optional[List[Model]]
    _iterations_count: int
    _forced_iteration: bool
    # Hold a lock since we are relying on internal state that circumvents standard next() behavior
    _lock: threading.Lock

    def __init__(
        self,
        api_call: PagedApiCall,
        results_extractor: ResultsExtractor,
        token_extractor: TokenExtractor = _default_token_extractor,
        length_extractor: LengthExtractor = _default_length_extractor,
    ) -> None:
        """
        Initialize a PageIterator.

        :param api_call: A Callable that accepts a str for the next_token
        :param results_extractor: A Callable that can extract a list of models from a results body
        :param token_extractor: A Callable that can extract the next_token str value from a results body
        :param length_extractor: A Callable that can extract the approximated available results
            count from a response
        """
        self._api_call = api_call
        self._results_extractor = results_extractor
        self._token_extractor = token_extractor
        self._length_extractor = length_extractor
        self._next_token = None
        self._approximate_length = None
        self._first_page = None
        self._iterations_count = 0
        self._forced_iteration = False
        self._lock = threading.Lock()

    def _fetch_data(self) -> List[Model]:
        with self._lock:
            # If we previously fetched the first page without the iterator being advanced, just return the
            # first page
            if self._forced_iteration and self._iterations_count > 0:
                self._forced_iteration = False
                # Appease MyPy with explicit [] return
                return self._first_page if self._first_page else []
            next_token = self._next_token if self._next_token else ""
            response: Response[Model] = self._api_call(next_token)
            result_body: Model = model_from_detailed(response)
            self._approximate_length = self._length_extractor(response)
            self._next_token = self._token_extractor(result_body)
            sdk_logger.debug(
                "Next token: %s Approximate Result Count: %s", self._next_token, self._approximate_length
            )
            data = self._results_extractor(result_body)
            results = [] if data is None else data
            if self._iterations_count == 0:
                self._first_page = results
            self._iterations_count += 1
            return results

    @property
    def next_token(self) -> Optional[NextToken]:
        """Return the last nextToken provided by the API, if applicable."""
        return self._next_token

    @property
    def estimated_count(self) -> int:
        """
        Return the value of the Result-Count header provided by the API, if applicable.

        Some endpoints may not implement this header. In that case, NotImplementedError is raised.
        """
        if self._first_page is None:
            self._forced_iteration = True
            self._first_page = self._fetch_data()
        if self._approximate_length is None:
            raise NotImplementedError("The API does not support Estimated Count for this operation")
        return self._approximate_length

    def first(self) -> Optional[Model]:
        """
        Return the first item from the first page.

        Will return None if the page had zero results.
        Operates independent of iteration - calls to `first` after starting iteration will still
        return the first item from the first page and not the current page.
        """
        if self._first_page is None:
            self._forced_iteration = True
            self._first_page = self._fetch_data()
        if self._first_page and len(self._first_page) > 0:
            return self._first_page[0]
        return None

    def __iter__(self) -> Iterator[List[Model]]:
        return self

    def __next__(self) -> List[Model]:
        if (
            self._next_token is None
            or len(self._next_token) > 0
            or (self._first_page is not None and self._iterations_count == 1 and self._forced_iteration)
        ):
            return self._fetch_data()
        raise StopIteration



from typing import Any, Dict, Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.molecules import (
    archive_molecules,
    bulk_create_molecules,
    bulk_update_molecules,
    bulk_upsert_molecules,
    create_molecule,
    get_molecule,
    list_molecules,
    unarchive_molecules,
    update_molecule,
    upsert_molecule,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import check_for_csv_bug_fix
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.models import (
    AsyncTaskLink,
    ListMoleculesSort,
    Molecule,
    MoleculeBulkUpdate,
    MoleculeCreate,
    MoleculesArchivalChange,
    MoleculesArchive,
    MoleculesArchiveReason,
    MoleculesBulkCreateRequest,
    MoleculesBulkUpdateRequest,
    MoleculesBulkUpsertRequest,
    MoleculesPaginatedList,
    MoleculesUnarchive,
    MoleculeUpdate,
    MoleculeUpsertRequest,
)
from benchling_sdk.services.v2.base_service import BaseService


class MoleculeService(BaseService):
    """
    Molecules.

    Molecules are groups of atoms held together by bonds, representing entities smaller than DNA
    Sequences and AA Sequences. Just like other entities, they support schemas, tags, and aliases.

    See https://benchling.com/api/reference#/Molecules
    """

    @api_method
    def get_by_id(self, molecule_id: str) -> Molecule:
        """
        Get a molecule.

        See https://benchling.com/api/reference#/Molecules/getMolecule
        """
        response = get_molecule.sync_detailed(client=self.client, molecule_id=molecule_id)
        return model_from_detailed(response)

    @api_method
    def _molecules_page(
        self,
        modified_at: Optional[str] = None,
        created_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        sort: Optional[ListMoleculesSort] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
        author_idsany_of: Optional[Iterable[str]] = None,
        chemical_substructuremol: Optional[str] = None,
        chemical_substructuresmiles: Optional[str] = None,
    ) -> Response[MoleculesPaginatedList]:
        response = list_molecules.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            created_at=none_as_unset(created_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            folder_id=none_as_unset(folder_id),
            mentioned_in=none_as_unset(mentioned_in),
            project_id=none_as_unset(project_id),
            registry_id=none_as_unset(registry_id),
            schema_id=none_as_unset(schema_id),
            archive_reason=none_as_unset(archive_reason),
            sort=none_as_unset(sort),
            ids=none_as_unset(optional_array_query_param(ids)),
            mentions=none_as_unset(mentions),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            entity_registry_idsany_of=none_as_unset(optional_array_query_param(entity_registry_ids_any_of)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            author_idsany_of=none_as_unset(optional_array_query_param(author_idsany_of)),
            chemical_substructuremol=none_as_unset(chemical_substructuremol),
            chemical_substructuresmiles=none_as_unset(chemical_substructuresmiles),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        modified_at: Optional[str] = None,
        created_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        sort: Optional[Union[str, ListMoleculesSort]] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
        author_idsany_of: Optional[Iterable[str]] = None,
        chemical_substructuremol: Optional[str] = None,
        chemical_substructuresmiles: Optional[str] = None,
    ) -> PageIterator[Molecule]:
        """
        List molecules.

        See https://benchling.com/api/reference#/Molecules/listMolecules
        """
        check_for_csv_bug_fix("mentioned_in", mentioned_in)
        check_for_csv_bug_fix("mentions", mentions)

        def api_call(next_token: NextToken) -> Response[MoleculesPaginatedList]:
            return self._molecules_page(
                modified_at=modified_at,
                created_at=created_at,
                name=name,
                name_includes=name_includes,
                folder_id=folder_id,
                mentioned_in=mentioned_in,
                project_id=project_id,
                registry_id=registry_id,
                schema_id=schema_id,
                archive_reason=archive_reason,
                mentions=mentions,
                ids=ids,
                entity_registry_ids_any_of=entity_registry_ids_any_of,
                names_any_of=names_any_of,
                schema_fields=schema_fields,
                sort=_translate_to_string_enum(ListMoleculesSort, sort),
                page_size=page_size,
                next_token=next_token,
                author_idsany_of=author_idsany_of,
                chemical_substructuremol=chemical_substructuremol,
                chemical_substructuresmiles=chemical_substructuresmiles,
            )

        def results_extractor(body: MoleculesPaginatedList) -> Optional[List[Molecule]]:
            return body.molecules

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, molecule: MoleculeCreate) -> Molecule:
        """
        Create a molecule.

        See https://benchling.com/api/reference#/Molecules/createMolecule
        """
        response = create_molecule.sync_detailed(client=self.client, json_body=molecule)
        return model_from_detailed(response)

    @api_method
    def update(self, molecule_id: str, molecule: MoleculeUpdate) -> Molecule:
        """
        Update a molecule.

        See https://benchling.com/api/reference#/Molecules/updateMolecule
        """
        response = update_molecule.sync_detailed(
            client=self.client, molecule_id=molecule_id, json_body=molecule
        )
        return model_from_detailed(response)

    @api_method
    def archive(self, molecule_ids: Iterable[str], reason: MoleculesArchiveReason) -> MoleculesArchivalChange:
        """
        Archive molecules.

        See https://benchling.com/api/reference#/Molecules/archiveMolecules
        """
        archive_request = MoleculesArchive(reason=reason, molecule_ids=list(molecule_ids))
        response = archive_molecules.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, molecule_ids: Iterable[str]) -> MoleculesArchivalChange:
        """
        Unarchive molecules.

        See https://benchling.com/api/reference#/Molecules/unarchiveMolecules
        """
        unarchive_request = MoleculesUnarchive(molecule_ids=list(molecule_ids))
        response = unarchive_molecules.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def bulk_create(self, molecules: Iterable[MoleculeCreate]) -> AsyncTaskLink:
        """
        Bulk create molecules.

        See https://benchling.com/api/reference#/Molecules/bulkCreateMolecules
        """
        body = MoleculesBulkCreateRequest(list(molecules))
        response = bulk_create_molecules.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update(self, molecules: Iterable[MoleculeBulkUpdate]) -> AsyncTaskLink:
        """
        Bulk update molecules.

        See https://benchling.com/api/reference#/Molecules/bulkUpdateMolecules
        """
        body = MoleculesBulkUpdateRequest(list(molecules))
        response = bulk_update_molecules.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def upsert(self, entity_registry_id: str, molecule: MoleculeUpsertRequest) -> Molecule:
        """
        Create or modify a Molecule.

        See https://benchling.com/api/reference#/Molecules/upsertMolecule
        """
        response = upsert_molecule.sync_detailed(
            client=self.client, entity_registry_id=entity_registry_id, json_body=molecule
        )
        return model_from_detailed(response)

    @api_method
    def bulk_upsert(
        self, body: MoleculesBulkUpsertRequest, returning: Optional[Iterable[str]] = None
    ) -> AsyncTaskLink:
        """
        Bulk create or update Molecules.

        See https://benchling.com/api/reference#/Molecules/bulkUpsertMolecules
        """
        returning_string = optional_array_query_param(returning)
        response = bulk_upsert_molecules.sync_detailed(
            client=self.client, json_body=body, returning=none_as_unset(returning_string)
        )
        return model_from_detailed(response)



from benchling_api_client.v2.stable.api.exports import export_item

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.task_helpers import TaskHelper
from benchling_sdk.models import ExportItemRequest, ExportsAsyncTaskResponse
from benchling_sdk.services.v2.base_service import BaseService


class ExportService(BaseService):
    """
    Exports.

    Export a Notebook Entry.

    See https://benchling.com/api/reference#/Exports
    """

    @api_method
    def export(self, export_request: ExportItemRequest) -> TaskHelper[ExportsAsyncTaskResponse]:
        """
        Export a Notebook Entry.

        This endpoint launches a long-running task and returns the Task ID of the launched task.
        The task response contains a link to download the exported item from Amazon S3.
        The download is a ZIP file that contains the exported PDFs.

        See https://benchling.com/api/reference#/Exports/exportItem
        """
        response = export_item.sync_detailed(client=self.client, json_body=export_request)
        return self._task_helper_from_response(response, ExportsAsyncTaskResponse)



from typing import Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.feature_libraries import (
    bulk_create_features,
    create_feature,
    create_feature_library,
    get_feature,
    get_feature_library,
    list_feature_libraries,
    list_features,
    update_feature,
    update_feature_library,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.helpers.task_helpers import TaskHelper
from benchling_sdk.models import (
    BulkCreateFeaturesAsyncTaskResponse,
    Feature,
    FeatureBulkCreate,
    FeatureCreate,
    FeatureLibrariesPaginatedList,
    FeatureLibrary,
    FeatureLibraryCreate,
    FeatureLibraryUpdate,
    FeaturesBulkCreateRequest,
    FeaturesPaginatedList,
    FeatureUpdate,
    ListFeatureLibrariesSort,
    ListFeaturesMatchType,
)
from benchling_sdk.services.v2.base_service import BaseService


class FeatureLibraryService(BaseService):
    """
    Feature Libraries.

    Feature Libraries are collections of shared canonical patterns that can be used to generate annotations
    on matching regions of DNA Sequences or AA Sequences.
    See https://benchling.com/api/reference#/Feature%20Libraries
    """

    @api_method
    def get_by_id(self, feature_library_id: str, returning: Optional[Iterable[str]] = None) -> FeatureLibrary:
        """
        Get a feature library by id.

        See https://benchling.com/api/reference#/Feature%20Libraries/getFeatureLibrary
        """
        returning_string = optional_array_query_param(returning)
        response = get_feature_library.sync_detailed(
            client=self.client,
            feature_library_id=feature_library_id,
            returning=none_as_unset(returning_string),
        )
        return model_from_detailed(response)

    @api_method
    def _feature_libraries_page(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        sort: Optional[ListFeatureLibrariesSort] = None,
        page_size: Optional[int] = None,
        returning: Optional[Iterable[str]] = None,
        next_token: Optional[NextToken] = None,
    ) -> Response[FeatureLibrariesPaginatedList]:
        response = list_feature_libraries.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            ids=none_as_unset(optional_array_query_param(ids)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            sort=none_as_unset(sort),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        sort: Optional[Union[str, ListFeatureLibrariesSort]] = None,
        page_size: Optional[int] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> PageIterator[FeatureLibrary]:
        """
        List Feature Libraries.

        See https://benchling.com/api/reference#/Feature%20Libraries/listFeatureLibraries
        """

        def api_call(next_token: NextToken) -> Response[FeatureLibrariesPaginatedList]:
            return self._feature_libraries_page(
                modified_at=modified_at,
                name=name,
                name_includes=name_includes,
                ids=ids,
                names_any_of=names_any_of,
                sort=_translate_to_string_enum(ListFeatureLibrariesSort, sort),
                page_size=page_size,
                next_token=next_token,
                returning=returning,
            )

        def results_extractor(body: FeatureLibrariesPaginatedList) -> Optional[List[FeatureLibrary]]:
            return body.feature_libraries

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, feature_library: FeatureLibraryCreate) -> FeatureLibrary:
        """
        Create a new feature library.

        See https://benchling.com/api/reference#/Feature%20Libraries/createFeatureLibrary
        """
        response = create_feature_library.sync_detailed(client=self.client, json_body=feature_library)
        return model_from_detailed(response)

    @api_method
    def update(self, feature_library_id: str, feature_library: FeatureLibraryUpdate) -> FeatureLibrary:
        """
        Update a feature library.

        See https://benchling.com/api/reference#/Feature%20Libraries/updateFeatureLibrary
        """
        response = update_feature_library.sync_detailed(
            client=self.client, feature_library_id=feature_library_id, json_body=feature_library
        )
        return model_from_detailed(response)

    # The below methods deal with features instead of feature libraries.

    @api_method
    def get_feature_by_id(self, feature_id: str, returning: Optional[Iterable[str]] = None) -> Feature:
        """
        Get a feature by id.

        See https://benchling.com/api/reference#/Feature%20Libraries/getFeature
        """
        returning_string = optional_array_query_param(returning)
        response = get_feature.sync_detailed(
            client=self.client, feature_id=feature_id, returning=none_as_unset(returning_string)
        )
        return model_from_detailed(response)

    @api_method
    def _features_page(
        self,
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
        name: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        namesany_ofcase_sensitive: Optional[Iterable[str]] = None,
        feature_library_id: Optional[str] = None,
        feature_type: Optional[str] = None,
        match_type: Optional[ListFeaturesMatchType] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> Response[FeaturesPaginatedList]:

        return list_features.sync_detailed(  # type: ignore
            client=self.client,
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            name=none_as_unset(name),
            ids=none_as_unset(optional_array_query_param(ids)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(namesany_ofcase_sensitive)),
            feature_library_id=none_as_unset(feature_library_id),
            feature_type=none_as_unset(feature_type),
            match_type=none_as_unset(match_type),
            returning=none_as_unset(optional_array_query_param(returning)),
        )

    def list_features(
        self,
        page_size: Optional[int] = None,
        name: Optional[str] = None,
        ids: Optional[Iterable[str]] = None,
        namesany_ofcase_sensitive: Optional[Iterable[str]] = None,
        feature_library_id: Optional[str] = None,
        feature_type: Optional[str] = None,
        match_type: Optional[ListFeaturesMatchType] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> PageIterator[Feature]:
        """
        List Features.

        See https://benchling.com/api/reference#/Feature%20Libraries/listFeatures
        """

        def api_call(next_token: NextToken) -> Response[FeaturesPaginatedList]:
            return self._features_page(
                page_size=page_size,
                next_token=next_token,
                name=name,
                ids=ids,
                namesany_ofcase_sensitive=namesany_ofcase_sensitive,
                feature_library_id=feature_library_id,
                feature_type=feature_type,
                match_type=match_type,
                returning=returning,
            )

        def results_extractor(body: FeaturesPaginatedList) -> Optional[List[Feature]]:
            return body.features

        return PageIterator(api_call, results_extractor)

    @api_method
    def create_feature(self, feature: FeatureCreate) -> Feature:
        """
        Create a feature.

        See https://benchling.com/api/reference#/Feature%20Libraries/createFeature
        """
        response = create_feature.sync_detailed(client=self.client, json_body=feature)
        return model_from_detailed(response)

    @api_method
    def update_feature(self, feature_id: str, feature: FeatureUpdate) -> Feature:
        """
        Update a feature.

        See https://benchling.com/api/reference#/Feature%20Libraries/updateFeature
        """
        response = update_feature.sync_detailed(client=self.client, feature_id=feature_id, json_body=feature)
        return model_from_detailed(response)

    @api_method
    def bulk_create_features(
        self, features: Iterable[FeatureBulkCreate]
    ) -> TaskHelper[BulkCreateFeaturesAsyncTaskResponse]:
        """
        Bulk create features.

        See https://benchling.com/api/reference#/Feature%20Libraries/bulkCreateFeatures
        """
        body = FeaturesBulkCreateRequest(list(features))
        response = bulk_create_features.sync_detailed(client=self.client, json_body=body)
        return self._task_helper_from_response(response, BulkCreateFeaturesAsyncTaskResponse)



# Re-export benchling_sdk.models benchling_sdk such that
# users can import from benchling_sdk without exposing benchling_api_client.
#
# Do not write by hand. Run `poetry run task models` to generate.
# This file should be committed as part of source control.


import sys
from typing import TYPE_CHECKING

__all__ = [
    "AppActivateRequestedWebhookV2",
    "AppActivateRequestedWebhookV2Type",
    "AppDeactivatedWebhookV2",
    "AppDeactivatedWebhookV2Type",
    "AppInstalledWebhookV2",
    "AppInstalledWebhookV2Type",
    "AssayRunCreatedWebhookV2",
    "AssayRunCreatedWebhookV2Type",
    "AssayRunUpdatedFieldsWebhookV2",
    "AssayRunUpdatedFieldsWebhookV2Type",
    "CanvasCreatedWebhookV2",
    "CanvasCreatedWebhookV2Beta",
    "CanvasCreatedWebhookV2BetaType",
    "CanvasCreatedWebhookV2Type",
    "CanvasInitializeWebhookV2",
    "CanvasInitializeWebhookV2Type",
    "CanvasInteractionWebhookV2",
    "CanvasInteractionWebhookV2Type",
    "EntityRegisteredWebhookV2",
    "EntityRegisteredWebhookV2Type",
    "EntryCreatedWebhookV2",
    "EntryCreatedWebhookV2Type",
    "EntryUpdatedFieldsWebhookV2",
    "EntryUpdatedFieldsWebhookV2Type",
    "EntryUpdatedReviewRecordWebhookV2",
    "EntryUpdatedReviewRecordWebhookV2Type",
    "EventBaseWebhookV2",
    "EventBaseWebhookV2Schematized",
    "EventResourceSchema",
    "LifecycleActivateWebhookV0",
    "LifecycleActivateWebhookV0Beta",
    "LifecycleActivateWebhookV0BetaType",
    "LifecycleActivateWebhookV0Type",
    "LifecycleConfigurationUpdateWebhookV0Beta",
    "LifecycleConfigurationUpdateWebhookV0BetaType",
    "LifecycleConfigurationUpdateWebhookV2Beta",
    "LifecycleConfigurationUpdateWebhookV2BetaType",
    "LifecycleDeactivateWebhookV0",
    "LifecycleDeactivateWebhookV0Beta",
    "LifecycleDeactivateWebhookV0BetaType",
    "LifecycleDeactivateWebhookV0Type",
    "MessageBase",
    "MessageBaseV0",
    "RequestCreatedWebhookV2",
    "RequestCreatedWebhookV2Type",
    "RequestUpdatedFieldsWebhookV2",
    "RequestUpdatedFieldsWebhookV2Type",
    "RequestUpdatedStatusWebhookV2",
    "RequestUpdatedStatusWebhookV2Type",
    "UpdateEventBaseWebhookV2",
    "UpdateEventBaseWebhookV2Schematized",
    "V2AssayRunCreatedEvent",
    "V2AssayRunCreatedEventEventType",
    "V2AssayRunUpdatedFieldsEvent",
    "V2AssayRunUpdatedFieldsEventEventType",
    "V2EntityRegisteredEvent",
    "V2EntityRegisteredEventEventType",
    "V2EntryCreatedEvent",
    "V2EntryCreatedEventEventType",
    "V2EntryUpdatedFieldsEvent",
    "V2EntryUpdatedFieldsEventEventType",
    "V2EntryUpdatedReviewRecordEvent",
    "V2EntryUpdatedReviewRecordEventEventType",
    "V2RequestCreatedEvent",
    "V2RequestCreatedEventEventType",
    "V2RequestUpdatedFieldsEvent",
    "V2RequestUpdatedFieldsEventEventType",
    "V2RequestUpdatedStatusEvent",
    "V2RequestUpdatedStatusEventEventType",
    "V2WorkflowOutputCreatedEvent",
    "V2WorkflowOutputCreatedEventEventType",
    "V2WorkflowOutputUpdatedFieldsEvent",
    "V2WorkflowOutputUpdatedFieldsEventEventType",
    "V2WorkflowTaskCreatedEvent",
    "V2WorkflowTaskCreatedEventEventType",
    "V2WorkflowTaskGroupCreatedEvent",
    "V2WorkflowTaskGroupCreatedEventEventType",
    "V2WorkflowTaskGroupUpdatedWatchersEvent",
    "V2WorkflowTaskGroupUpdatedWatchersEventEventType",
    "V2WorkflowTaskUpdatedAssigneeEvent",
    "V2WorkflowTaskUpdatedAssigneeEventEventType",
    "V2WorkflowTaskUpdatedFieldsEvent",
    "V2WorkflowTaskUpdatedFieldsEventEventType",
    "V2WorkflowTaskUpdatedScheduledOnEvent",
    "V2WorkflowTaskUpdatedScheduledOnEventEventType",
    "V2WorkflowTaskUpdatedStatusEvent",
    "V2WorkflowTaskUpdatedStatusEventEventType",
    "WebhookEnvelopeV0",
    "WebhookEnvelopeV0App",
    "WebhookEnvelopeV0AppDefinition",
    "WebhookEnvelopeV0Version",
    "WebhookMessageV0",
    "WorkflowOutputCreatedWebhookV2",
    "WorkflowOutputCreatedWebhookV2Type",
    "WorkflowOutputUpdatedFieldsWebhookV2",
    "WorkflowOutputUpdatedFieldsWebhookV2Type",
    "WorkflowTaskCreatedWebhookV2",
    "WorkflowTaskCreatedWebhookV2Type",
    "WorkflowTaskGroupCreatedWebhookV2",
    "WorkflowTaskGroupCreatedWebhookV2Type",
    "WorkflowTaskGroupMappingCompletedWebhookV2",
    "WorkflowTaskGroupMappingCompletedWebhookV2Type",
    "WorkflowTaskGroupUpdatedWatchersWebhookV2",
    "WorkflowTaskGroupUpdatedWatchersWebhookV2Type",
    "WorkflowTaskUpdatedAssigneeWebhookV2",
    "WorkflowTaskUpdatedAssigneeWebhookV2Type",
    "WorkflowTaskUpdatedFieldsWebhookV2",
    "WorkflowTaskUpdatedFieldsWebhookV2Type",
    "WorkflowTaskUpdatedScheduledOnWebhookV2",
    "WorkflowTaskUpdatedScheduledOnWebhookV2Type",
    "WorkflowTaskUpdatedStatusWebhookV2",
    "WorkflowTaskUpdatedStatusWebhookV2Type",
]

if TYPE_CHECKING:
    import benchling_api_client.webhooks.v0.stable.models.app_activate_requested_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.app_activate_requested_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.app_deactivated_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.app_deactivated_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.app_installed_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.app_installed_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.assay_run_created_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.assay_run_created_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.assay_run_updated_fields_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.assay_run_updated_fields_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2_beta
    import benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2_beta_type
    import benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.canvas_initialize_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.canvas_initialize_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.canvas_interaction_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.canvas_interaction_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.entity_registered_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.entity_registered_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.entry_created_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.entry_created_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.entry_updated_fields_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.entry_updated_fields_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.entry_updated_review_record_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.entry_updated_review_record_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.event_base_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.event_base_webhook_v2_schematized
    import benchling_api_client.webhooks.v0.stable.models.event_resource_schema
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0_beta
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0_beta_type
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0_type
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v0_beta
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v0_beta_type
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v2_beta
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v2_beta_type
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0_beta
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0_beta_type
    import benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0_type
    import benchling_api_client.webhooks.v0.stable.models.message_base
    import benchling_api_client.webhooks.v0.stable.models.message_base_v0
    import benchling_api_client.webhooks.v0.stable.models.request_created_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.request_created_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.request_updated_fields_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.request_updated_fields_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.request_updated_status_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.request_updated_status_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.update_event_base_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.update_event_base_webhook_v2_schematized
    import benchling_api_client.webhooks.v0.stable.models.v2_assay_run_created_event
    import benchling_api_client.webhooks.v0.stable.models.v2_assay_run_created_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_assay_run_updated_fields_event
    import benchling_api_client.webhooks.v0.stable.models.v2_assay_run_updated_fields_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_entity_registered_event
    import benchling_api_client.webhooks.v0.stable.models.v2_entity_registered_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_entry_created_event
    import benchling_api_client.webhooks.v0.stable.models.v2_entry_created_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_fields_event
    import benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_fields_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_review_record_event
    import benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_review_record_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_request_created_event
    import benchling_api_client.webhooks.v0.stable.models.v2_request_created_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_request_updated_fields_event
    import benchling_api_client.webhooks.v0.stable.models.v2_request_updated_fields_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_request_updated_status_event
    import benchling_api_client.webhooks.v0.stable.models.v2_request_updated_status_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_created_event
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_created_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_updated_fields_event
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_updated_fields_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_created_event
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_created_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_created_event
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_created_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_updated_watchers_event
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_updated_watchers_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_assignee_event
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_assignee_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_fields_event
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_fields_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_scheduled_on_event
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_scheduled_on_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_status_event
    import benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_status_event_event_type
    import benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0
    import benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0_app
    import benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0_app_definition
    import benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0_version
    import benchling_api_client.webhooks.v0.stable.models.webhook_message_v0
    import benchling_api_client.webhooks.v0.stable.models.workflow_output_created_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.workflow_output_created_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.workflow_output_updated_fields_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.workflow_output_updated_fields_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_created_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_created_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_group_created_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_group_created_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_group_mapping_completed_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_group_mapping_completed_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_group_updated_watchers_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_group_updated_watchers_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_assignee_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_assignee_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_fields_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_fields_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_scheduled_on_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_scheduled_on_webhook_v2_type
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_status_webhook_v2
    import benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_status_webhook_v2_type

    AppActivateRequestedWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.app_activate_requested_webhook_v2.AppActivateRequestedWebhookV2
    )
    AppActivateRequestedWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.app_activate_requested_webhook_v2_type.AppActivateRequestedWebhookV2Type
    )
    AppDeactivatedWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.app_deactivated_webhook_v2.AppDeactivatedWebhookV2
    )
    AppDeactivatedWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.app_deactivated_webhook_v2_type.AppDeactivatedWebhookV2Type
    )
    AppInstalledWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.app_installed_webhook_v2.AppInstalledWebhookV2
    )
    AppInstalledWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.app_installed_webhook_v2_type.AppInstalledWebhookV2Type
    )
    AssayRunCreatedWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.assay_run_created_webhook_v2.AssayRunCreatedWebhookV2
    )
    AssayRunCreatedWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.assay_run_created_webhook_v2_type.AssayRunCreatedWebhookV2Type
    )
    AssayRunUpdatedFieldsWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.assay_run_updated_fields_webhook_v2.AssayRunUpdatedFieldsWebhookV2
    )
    AssayRunUpdatedFieldsWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.assay_run_updated_fields_webhook_v2_type.AssayRunUpdatedFieldsWebhookV2Type
    )
    CanvasCreatedWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2.CanvasCreatedWebhookV2
    )
    CanvasCreatedWebhookV2Beta = (
        benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2_beta.CanvasCreatedWebhookV2Beta
    )
    CanvasCreatedWebhookV2BetaType = (
        benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2_beta_type.CanvasCreatedWebhookV2BetaType
    )
    CanvasCreatedWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2_type.CanvasCreatedWebhookV2Type
    )
    CanvasInitializeWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.canvas_initialize_webhook_v2.CanvasInitializeWebhookV2
    )
    CanvasInitializeWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.canvas_initialize_webhook_v2_type.CanvasInitializeWebhookV2Type
    )
    CanvasInteractionWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.canvas_interaction_webhook_v2.CanvasInteractionWebhookV2
    )
    CanvasInteractionWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.canvas_interaction_webhook_v2_type.CanvasInteractionWebhookV2Type
    )
    EntityRegisteredWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.entity_registered_webhook_v2.EntityRegisteredWebhookV2
    )
    EntityRegisteredWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.entity_registered_webhook_v2_type.EntityRegisteredWebhookV2Type
    )
    EntryCreatedWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.entry_created_webhook_v2.EntryCreatedWebhookV2
    )
    EntryCreatedWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.entry_created_webhook_v2_type.EntryCreatedWebhookV2Type
    )
    EntryUpdatedFieldsWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.entry_updated_fields_webhook_v2.EntryUpdatedFieldsWebhookV2
    )
    EntryUpdatedFieldsWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.entry_updated_fields_webhook_v2_type.EntryUpdatedFieldsWebhookV2Type
    )
    EntryUpdatedReviewRecordWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.entry_updated_review_record_webhook_v2.EntryUpdatedReviewRecordWebhookV2
    )
    EntryUpdatedReviewRecordWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.entry_updated_review_record_webhook_v2_type.EntryUpdatedReviewRecordWebhookV2Type
    )
    EventBaseWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.event_base_webhook_v2.EventBaseWebhookV2
    )
    EventBaseWebhookV2Schematized = (
        benchling_api_client.webhooks.v0.stable.models.event_base_webhook_v2_schematized.EventBaseWebhookV2Schematized
    )
    EventResourceSchema = (
        benchling_api_client.webhooks.v0.stable.models.event_resource_schema.EventResourceSchema
    )
    LifecycleActivateWebhookV0 = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0.LifecycleActivateWebhookV0
    )
    LifecycleActivateWebhookV0Beta = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0_beta.LifecycleActivateWebhookV0Beta
    )
    LifecycleActivateWebhookV0BetaType = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0_beta_type.LifecycleActivateWebhookV0BetaType
    )
    LifecycleActivateWebhookV0Type = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0_type.LifecycleActivateWebhookV0Type
    )
    LifecycleConfigurationUpdateWebhookV0Beta = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v0_beta.LifecycleConfigurationUpdateWebhookV0Beta
    )
    LifecycleConfigurationUpdateWebhookV0BetaType = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v0_beta_type.LifecycleConfigurationUpdateWebhookV0BetaType
    )
    LifecycleConfigurationUpdateWebhookV2Beta = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v2_beta.LifecycleConfigurationUpdateWebhookV2Beta
    )
    LifecycleConfigurationUpdateWebhookV2BetaType = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v2_beta_type.LifecycleConfigurationUpdateWebhookV2BetaType
    )
    LifecycleDeactivateWebhookV0 = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0.LifecycleDeactivateWebhookV0
    )
    LifecycleDeactivateWebhookV0Beta = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0_beta.LifecycleDeactivateWebhookV0Beta
    )
    LifecycleDeactivateWebhookV0BetaType = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0_beta_type.LifecycleDeactivateWebhookV0BetaType
    )
    LifecycleDeactivateWebhookV0Type = (
        benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0_type.LifecycleDeactivateWebhookV0Type
    )
    MessageBase = (
        benchling_api_client.webhooks.v0.stable.models.message_base.MessageBase
    )
    MessageBaseV0 = (
        benchling_api_client.webhooks.v0.stable.models.message_base_v0.MessageBaseV0
    )
    RequestCreatedWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.request_created_webhook_v2.RequestCreatedWebhookV2
    )
    RequestCreatedWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.request_created_webhook_v2_type.RequestCreatedWebhookV2Type
    )
    RequestUpdatedFieldsWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.request_updated_fields_webhook_v2.RequestUpdatedFieldsWebhookV2
    )
    RequestUpdatedFieldsWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.request_updated_fields_webhook_v2_type.RequestUpdatedFieldsWebhookV2Type
    )
    RequestUpdatedStatusWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.request_updated_status_webhook_v2.RequestUpdatedStatusWebhookV2
    )
    RequestUpdatedStatusWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.request_updated_status_webhook_v2_type.RequestUpdatedStatusWebhookV2Type
    )
    UpdateEventBaseWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.update_event_base_webhook_v2.UpdateEventBaseWebhookV2
    )
    UpdateEventBaseWebhookV2Schematized = (
        benchling_api_client.webhooks.v0.stable.models.update_event_base_webhook_v2_schematized.UpdateEventBaseWebhookV2Schematized
    )
    V2AssayRunCreatedEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_assay_run_created_event.V2AssayRunCreatedEvent
    )
    V2AssayRunCreatedEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_assay_run_created_event_event_type.V2AssayRunCreatedEventEventType
    )
    V2AssayRunUpdatedFieldsEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_assay_run_updated_fields_event.V2AssayRunUpdatedFieldsEvent
    )
    V2AssayRunUpdatedFieldsEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_assay_run_updated_fields_event_event_type.V2AssayRunUpdatedFieldsEventEventType
    )
    V2EntityRegisteredEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_entity_registered_event.V2EntityRegisteredEvent
    )
    V2EntityRegisteredEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_entity_registered_event_event_type.V2EntityRegisteredEventEventType
    )
    V2EntryCreatedEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_entry_created_event.V2EntryCreatedEvent
    )
    V2EntryCreatedEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_entry_created_event_event_type.V2EntryCreatedEventEventType
    )
    V2EntryUpdatedFieldsEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_fields_event.V2EntryUpdatedFieldsEvent
    )
    V2EntryUpdatedFieldsEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_fields_event_event_type.V2EntryUpdatedFieldsEventEventType
    )
    V2EntryUpdatedReviewRecordEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_review_record_event.V2EntryUpdatedReviewRecordEvent
    )
    V2EntryUpdatedReviewRecordEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_review_record_event_event_type.V2EntryUpdatedReviewRecordEventEventType
    )
    V2RequestCreatedEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_request_created_event.V2RequestCreatedEvent
    )
    V2RequestCreatedEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_request_created_event_event_type.V2RequestCreatedEventEventType
    )
    V2RequestUpdatedFieldsEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_request_updated_fields_event.V2RequestUpdatedFieldsEvent
    )
    V2RequestUpdatedFieldsEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_request_updated_fields_event_event_type.V2RequestUpdatedFieldsEventEventType
    )
    V2RequestUpdatedStatusEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_request_updated_status_event.V2RequestUpdatedStatusEvent
    )
    V2RequestUpdatedStatusEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_request_updated_status_event_event_type.V2RequestUpdatedStatusEventEventType
    )
    V2WorkflowOutputCreatedEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_created_event.V2WorkflowOutputCreatedEvent
    )
    V2WorkflowOutputCreatedEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_created_event_event_type.V2WorkflowOutputCreatedEventEventType
    )
    V2WorkflowOutputUpdatedFieldsEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_updated_fields_event.V2WorkflowOutputUpdatedFieldsEvent
    )
    V2WorkflowOutputUpdatedFieldsEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_updated_fields_event_event_type.V2WorkflowOutputUpdatedFieldsEventEventType
    )
    V2WorkflowTaskCreatedEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_created_event.V2WorkflowTaskCreatedEvent
    )
    V2WorkflowTaskCreatedEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_created_event_event_type.V2WorkflowTaskCreatedEventEventType
    )
    V2WorkflowTaskGroupCreatedEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_created_event.V2WorkflowTaskGroupCreatedEvent
    )
    V2WorkflowTaskGroupCreatedEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_created_event_event_type.V2WorkflowTaskGroupCreatedEventEventType
    )
    V2WorkflowTaskGroupUpdatedWatchersEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_updated_watchers_event.V2WorkflowTaskGroupUpdatedWatchersEvent
    )
    V2WorkflowTaskGroupUpdatedWatchersEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_updated_watchers_event_event_type.V2WorkflowTaskGroupUpdatedWatchersEventEventType
    )
    V2WorkflowTaskUpdatedAssigneeEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_assignee_event.V2WorkflowTaskUpdatedAssigneeEvent
    )
    V2WorkflowTaskUpdatedAssigneeEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_assignee_event_event_type.V2WorkflowTaskUpdatedAssigneeEventEventType
    )
    V2WorkflowTaskUpdatedFieldsEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_fields_event.V2WorkflowTaskUpdatedFieldsEvent
    )
    V2WorkflowTaskUpdatedFieldsEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_fields_event_event_type.V2WorkflowTaskUpdatedFieldsEventEventType
    )
    V2WorkflowTaskUpdatedScheduledOnEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_scheduled_on_event.V2WorkflowTaskUpdatedScheduledOnEvent
    )
    V2WorkflowTaskUpdatedScheduledOnEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_scheduled_on_event_event_type.V2WorkflowTaskUpdatedScheduledOnEventEventType
    )
    V2WorkflowTaskUpdatedStatusEvent = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_status_event.V2WorkflowTaskUpdatedStatusEvent
    )
    V2WorkflowTaskUpdatedStatusEventEventType = (
        benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_status_event_event_type.V2WorkflowTaskUpdatedStatusEventEventType
    )
    WebhookEnvelopeV0 = (
        benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0.WebhookEnvelopeV0
    )
    WebhookEnvelopeV0App = (
        benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0_app.WebhookEnvelopeV0App
    )
    WebhookEnvelopeV0AppDefinition = (
        benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0_app_definition.WebhookEnvelopeV0AppDefinition
    )
    WebhookEnvelopeV0Version = (
        benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0_version.WebhookEnvelopeV0Version
    )
    WebhookMessageV0 = (
        benchling_api_client.webhooks.v0.stable.models.webhook_message_v0.WebhookMessageV0
    )
    WorkflowOutputCreatedWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.workflow_output_created_webhook_v2.WorkflowOutputCreatedWebhookV2
    )
    WorkflowOutputCreatedWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.workflow_output_created_webhook_v2_type.WorkflowOutputCreatedWebhookV2Type
    )
    WorkflowOutputUpdatedFieldsWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.workflow_output_updated_fields_webhook_v2.WorkflowOutputUpdatedFieldsWebhookV2
    )
    WorkflowOutputUpdatedFieldsWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.workflow_output_updated_fields_webhook_v2_type.WorkflowOutputUpdatedFieldsWebhookV2Type
    )
    WorkflowTaskCreatedWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_created_webhook_v2.WorkflowTaskCreatedWebhookV2
    )
    WorkflowTaskCreatedWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_created_webhook_v2_type.WorkflowTaskCreatedWebhookV2Type
    )
    WorkflowTaskGroupCreatedWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_group_created_webhook_v2.WorkflowTaskGroupCreatedWebhookV2
    )
    WorkflowTaskGroupCreatedWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_group_created_webhook_v2_type.WorkflowTaskGroupCreatedWebhookV2Type
    )
    WorkflowTaskGroupMappingCompletedWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_group_mapping_completed_webhook_v2.WorkflowTaskGroupMappingCompletedWebhookV2
    )
    WorkflowTaskGroupMappingCompletedWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_group_mapping_completed_webhook_v2_type.WorkflowTaskGroupMappingCompletedWebhookV2Type
    )
    WorkflowTaskGroupUpdatedWatchersWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_group_updated_watchers_webhook_v2.WorkflowTaskGroupUpdatedWatchersWebhookV2
    )
    WorkflowTaskGroupUpdatedWatchersWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_group_updated_watchers_webhook_v2_type.WorkflowTaskGroupUpdatedWatchersWebhookV2Type
    )
    WorkflowTaskUpdatedAssigneeWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_assignee_webhook_v2.WorkflowTaskUpdatedAssigneeWebhookV2
    )
    WorkflowTaskUpdatedAssigneeWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_assignee_webhook_v2_type.WorkflowTaskUpdatedAssigneeWebhookV2Type
    )
    WorkflowTaskUpdatedFieldsWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_fields_webhook_v2.WorkflowTaskUpdatedFieldsWebhookV2
    )
    WorkflowTaskUpdatedFieldsWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_fields_webhook_v2_type.WorkflowTaskUpdatedFieldsWebhookV2Type
    )
    WorkflowTaskUpdatedScheduledOnWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_scheduled_on_webhook_v2.WorkflowTaskUpdatedScheduledOnWebhookV2
    )
    WorkflowTaskUpdatedScheduledOnWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_scheduled_on_webhook_v2_type.WorkflowTaskUpdatedScheduledOnWebhookV2Type
    )
    WorkflowTaskUpdatedStatusWebhookV2 = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_status_webhook_v2.WorkflowTaskUpdatedStatusWebhookV2
    )
    WorkflowTaskUpdatedStatusWebhookV2Type = (
        benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_status_webhook_v2_type.WorkflowTaskUpdatedStatusWebhookV2Type
    )

else:
    model_to_module_mapping = {
        "AppActivateRequestedWebhookV2": "benchling_api_client.webhooks.v0.stable.models.app_activate_requested_webhook_v2",
        "AppActivateRequestedWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.app_activate_requested_webhook_v2_type",
        "AppDeactivatedWebhookV2": "benchling_api_client.webhooks.v0.stable.models.app_deactivated_webhook_v2",
        "AppDeactivatedWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.app_deactivated_webhook_v2_type",
        "AppInstalledWebhookV2": "benchling_api_client.webhooks.v0.stable.models.app_installed_webhook_v2",
        "AppInstalledWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.app_installed_webhook_v2_type",
        "AssayRunCreatedWebhookV2": "benchling_api_client.webhooks.v0.stable.models.assay_run_created_webhook_v2",
        "AssayRunCreatedWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.assay_run_created_webhook_v2_type",
        "AssayRunUpdatedFieldsWebhookV2": "benchling_api_client.webhooks.v0.stable.models.assay_run_updated_fields_webhook_v2",
        "AssayRunUpdatedFieldsWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.assay_run_updated_fields_webhook_v2_type",
        "CanvasCreatedWebhookV2": "benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2",
        "CanvasCreatedWebhookV2Beta": "benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2_beta",
        "CanvasCreatedWebhookV2BetaType": "benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2_beta_type",
        "CanvasCreatedWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.canvas_created_webhook_v2_type",
        "CanvasInitializeWebhookV2": "benchling_api_client.webhooks.v0.stable.models.canvas_initialize_webhook_v2",
        "CanvasInitializeWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.canvas_initialize_webhook_v2_type",
        "CanvasInteractionWebhookV2": "benchling_api_client.webhooks.v0.stable.models.canvas_interaction_webhook_v2",
        "CanvasInteractionWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.canvas_interaction_webhook_v2_type",
        "EntityRegisteredWebhookV2": "benchling_api_client.webhooks.v0.stable.models.entity_registered_webhook_v2",
        "EntityRegisteredWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.entity_registered_webhook_v2_type",
        "EntryCreatedWebhookV2": "benchling_api_client.webhooks.v0.stable.models.entry_created_webhook_v2",
        "EntryCreatedWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.entry_created_webhook_v2_type",
        "EntryUpdatedFieldsWebhookV2": "benchling_api_client.webhooks.v0.stable.models.entry_updated_fields_webhook_v2",
        "EntryUpdatedFieldsWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.entry_updated_fields_webhook_v2_type",
        "EntryUpdatedReviewRecordWebhookV2": "benchling_api_client.webhooks.v0.stable.models.entry_updated_review_record_webhook_v2",
        "EntryUpdatedReviewRecordWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.entry_updated_review_record_webhook_v2_type",
        "EventBaseWebhookV2": "benchling_api_client.webhooks.v0.stable.models.event_base_webhook_v2",
        "EventBaseWebhookV2Schematized": "benchling_api_client.webhooks.v0.stable.models.event_base_webhook_v2_schematized",
        "EventResourceSchema": "benchling_api_client.webhooks.v0.stable.models.event_resource_schema",
        "LifecycleActivateWebhookV0": "benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0",
        "LifecycleActivateWebhookV0Beta": "benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0_beta",
        "LifecycleActivateWebhookV0BetaType": "benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0_beta_type",
        "LifecycleActivateWebhookV0Type": "benchling_api_client.webhooks.v0.stable.models.lifecycle_activate_webhook_v0_type",
        "LifecycleConfigurationUpdateWebhookV0Beta": "benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v0_beta",
        "LifecycleConfigurationUpdateWebhookV0BetaType": "benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v0_beta_type",
        "LifecycleConfigurationUpdateWebhookV2Beta": "benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v2_beta",
        "LifecycleConfigurationUpdateWebhookV2BetaType": "benchling_api_client.webhooks.v0.stable.models.lifecycle_configuration_update_webhook_v2_beta_type",
        "LifecycleDeactivateWebhookV0": "benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0",
        "LifecycleDeactivateWebhookV0Beta": "benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0_beta",
        "LifecycleDeactivateWebhookV0BetaType": "benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0_beta_type",
        "LifecycleDeactivateWebhookV0Type": "benchling_api_client.webhooks.v0.stable.models.lifecycle_deactivate_webhook_v0_type",
        "MessageBase": "benchling_api_client.webhooks.v0.stable.models.message_base",
        "MessageBaseV0": "benchling_api_client.webhooks.v0.stable.models.message_base_v0",
        "RequestCreatedWebhookV2": "benchling_api_client.webhooks.v0.stable.models.request_created_webhook_v2",
        "RequestCreatedWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.request_created_webhook_v2_type",
        "RequestUpdatedFieldsWebhookV2": "benchling_api_client.webhooks.v0.stable.models.request_updated_fields_webhook_v2",
        "RequestUpdatedFieldsWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.request_updated_fields_webhook_v2_type",
        "RequestUpdatedStatusWebhookV2": "benchling_api_client.webhooks.v0.stable.models.request_updated_status_webhook_v2",
        "RequestUpdatedStatusWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.request_updated_status_webhook_v2_type",
        "UpdateEventBaseWebhookV2": "benchling_api_client.webhooks.v0.stable.models.update_event_base_webhook_v2",
        "UpdateEventBaseWebhookV2Schematized": "benchling_api_client.webhooks.v0.stable.models.update_event_base_webhook_v2_schematized",
        "V2AssayRunCreatedEvent": "benchling_api_client.webhooks.v0.stable.models.v2_assay_run_created_event",
        "V2AssayRunCreatedEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_assay_run_created_event_event_type",
        "V2AssayRunUpdatedFieldsEvent": "benchling_api_client.webhooks.v0.stable.models.v2_assay_run_updated_fields_event",
        "V2AssayRunUpdatedFieldsEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_assay_run_updated_fields_event_event_type",
        "V2EntityRegisteredEvent": "benchling_api_client.webhooks.v0.stable.models.v2_entity_registered_event",
        "V2EntityRegisteredEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_entity_registered_event_event_type",
        "V2EntryCreatedEvent": "benchling_api_client.webhooks.v0.stable.models.v2_entry_created_event",
        "V2EntryCreatedEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_entry_created_event_event_type",
        "V2EntryUpdatedFieldsEvent": "benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_fields_event",
        "V2EntryUpdatedFieldsEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_fields_event_event_type",
        "V2EntryUpdatedReviewRecordEvent": "benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_review_record_event",
        "V2EntryUpdatedReviewRecordEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_entry_updated_review_record_event_event_type",
        "V2RequestCreatedEvent": "benchling_api_client.webhooks.v0.stable.models.v2_request_created_event",
        "V2RequestCreatedEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_request_created_event_event_type",
        "V2RequestUpdatedFieldsEvent": "benchling_api_client.webhooks.v0.stable.models.v2_request_updated_fields_event",
        "V2RequestUpdatedFieldsEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_request_updated_fields_event_event_type",
        "V2RequestUpdatedStatusEvent": "benchling_api_client.webhooks.v0.stable.models.v2_request_updated_status_event",
        "V2RequestUpdatedStatusEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_request_updated_status_event_event_type",
        "V2WorkflowOutputCreatedEvent": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_created_event",
        "V2WorkflowOutputCreatedEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_created_event_event_type",
        "V2WorkflowOutputUpdatedFieldsEvent": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_updated_fields_event",
        "V2WorkflowOutputUpdatedFieldsEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_output_updated_fields_event_event_type",
        "V2WorkflowTaskCreatedEvent": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_created_event",
        "V2WorkflowTaskCreatedEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_created_event_event_type",
        "V2WorkflowTaskGroupCreatedEvent": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_created_event",
        "V2WorkflowTaskGroupCreatedEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_created_event_event_type",
        "V2WorkflowTaskGroupUpdatedWatchersEvent": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_updated_watchers_event",
        "V2WorkflowTaskGroupUpdatedWatchersEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_group_updated_watchers_event_event_type",
        "V2WorkflowTaskUpdatedAssigneeEvent": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_assignee_event",
        "V2WorkflowTaskUpdatedAssigneeEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_assignee_event_event_type",
        "V2WorkflowTaskUpdatedFieldsEvent": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_fields_event",
        "V2WorkflowTaskUpdatedFieldsEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_fields_event_event_type",
        "V2WorkflowTaskUpdatedScheduledOnEvent": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_scheduled_on_event",
        "V2WorkflowTaskUpdatedScheduledOnEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_scheduled_on_event_event_type",
        "V2WorkflowTaskUpdatedStatusEvent": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_status_event",
        "V2WorkflowTaskUpdatedStatusEventEventType": "benchling_api_client.webhooks.v0.stable.models.v2_workflow_task_updated_status_event_event_type",
        "WebhookEnvelopeV0": "benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0",
        "WebhookEnvelopeV0App": "benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0_app",
        "WebhookEnvelopeV0AppDefinition": "benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0_app_definition",
        "WebhookEnvelopeV0Version": "benchling_api_client.webhooks.v0.stable.models.webhook_envelope_v0_version",
        "WebhookMessageV0": "benchling_api_client.webhooks.v0.stable.models.webhook_message_v0",
        "WorkflowOutputCreatedWebhookV2": "benchling_api_client.webhooks.v0.stable.models.workflow_output_created_webhook_v2",
        "WorkflowOutputCreatedWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.workflow_output_created_webhook_v2_type",
        "WorkflowOutputUpdatedFieldsWebhookV2": "benchling_api_client.webhooks.v0.stable.models.workflow_output_updated_fields_webhook_v2",
        "WorkflowOutputUpdatedFieldsWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.workflow_output_updated_fields_webhook_v2_type",
        "WorkflowTaskCreatedWebhookV2": "benchling_api_client.webhooks.v0.stable.models.workflow_task_created_webhook_v2",
        "WorkflowTaskCreatedWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.workflow_task_created_webhook_v2_type",
        "WorkflowTaskGroupCreatedWebhookV2": "benchling_api_client.webhooks.v0.stable.models.workflow_task_group_created_webhook_v2",
        "WorkflowTaskGroupCreatedWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.workflow_task_group_created_webhook_v2_type",
        "WorkflowTaskGroupMappingCompletedWebhookV2": "benchling_api_client.webhooks.v0.stable.models.workflow_task_group_mapping_completed_webhook_v2",
        "WorkflowTaskGroupMappingCompletedWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.workflow_task_group_mapping_completed_webhook_v2_type",
        "WorkflowTaskGroupUpdatedWatchersWebhookV2": "benchling_api_client.webhooks.v0.stable.models.workflow_task_group_updated_watchers_webhook_v2",
        "WorkflowTaskGroupUpdatedWatchersWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.workflow_task_group_updated_watchers_webhook_v2_type",
        "WorkflowTaskUpdatedAssigneeWebhookV2": "benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_assignee_webhook_v2",
        "WorkflowTaskUpdatedAssigneeWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_assignee_webhook_v2_type",
        "WorkflowTaskUpdatedFieldsWebhookV2": "benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_fields_webhook_v2",
        "WorkflowTaskUpdatedFieldsWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_fields_webhook_v2_type",
        "WorkflowTaskUpdatedScheduledOnWebhookV2": "benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_scheduled_on_webhook_v2",
        "WorkflowTaskUpdatedScheduledOnWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_scheduled_on_webhook_v2_type",
        "WorkflowTaskUpdatedStatusWebhookV2": "benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_status_webhook_v2",
        "WorkflowTaskUpdatedStatusWebhookV2Type": "benchling_api_client.webhooks.v0.stable.models.workflow_task_updated_status_webhook_v2_type",
    }

    from types import ModuleType

    # Custom module to allow for lazy loading of models
    class _Models(ModuleType):
        def __getattr__(self, name):
            if module_name := model_to_module_mapping.get(name):
                module = __import__(module_name, None, None, [name])
                setattr(self, name, getattr(module, name))
                return ModuleType.__getattribute__(self, name)
            return ModuleType.__getattr__(self, name)

    # keep a reference to this module so that it's not garbage collected
    old_module = sys.modules[__name__]

    new_module = sys.modules[__name__] = _Models(__name__)
    new_module.__dict__.update(
        {
            "__file__": __file__,
            "__path__": __path__,
            "__doc__": __doc__,
            "__all__": __all__,
        }
    )
